from __future__ import annotations

import json
import string
from enum import StrEnum
from hashlib import sha256
from typing import Annotated, Any, Dict, List

import httpx
from pydantic import BaseModel, BeforeValidator, Field, PlainSerializer, WithJsonSchema, model_validator

from .coretypes import ResourceId
from .property import PropertyKey
from .resource_abc import CamelAlias, Ref, Resource, register_resource
from .urlfmt import Urlfmt

url: string.Formatter = Urlfmt("api/api/datamodel")
DATA_MODEL_URL_FORMAT = "{base}/{entity_type}"
FULL_DATA_MODEL_URL = DATA_MODEL_URL_FORMAT + "/{scope}/{code}"


class SortOrder(StrEnum):
    ASC = "ASC"
    DESC = "DESC"


class DataModelProperty(CamelAlias, BaseModel):
    property_key: PropertyKey
    required: bool | None


class IdentifierType(CamelAlias, BaseModel):
    identifier_key: str
    required: bool | None


class AttributeAlias(CamelAlias, BaseModel):
    attribute_name: str
    attribute_alias: str


class RecommendedSortBy(CamelAlias, BaseModel):
    attribute_name: str
    sort_order: SortOrder | None


def ser_parent_data_model(value, info):
    """Serialize parent data model reference."""
    if info.context and info.context.get("style", "api") == "dump":
        if isinstance(value, (Resource, Ref)):
            return {"$ref": value.id}
        return value
    if isinstance(value, (CustomDataModelResource, CustomDataModelRef)):
        return {"scope": value.resource_id.scope, "code": value.resource_id.code}
    return value


def des_parent_data_model(value, info):
    """Deserialize parent data model reference."""
    if not isinstance(value, dict):
        return value
    if info.context and info.context.get("$refs") and "$ref" in value:
        ref = info.context["$refs"][value["$ref"]]
        return ref
    # Handle ResourceId format
    if "scope" in value and "code" in value:
        return value
    return value


ParentDataModelKey = Annotated[
    "CustomDataModelResource | CustomDataModelRef | ResourceId | None",
    BeforeValidator(des_parent_data_model),
    PlainSerializer(ser_parent_data_model),
    WithJsonSchema(
        {
            "type": "object",
            "properties": {"$ref": {"type": "string", "format": "Key.CustomDataModel"}},
            "required": ["$ref"],
        }
    ),
]


@register_resource()
class CustomDataModelRef(CamelAlias, BaseModel, Ref):
    """Reference to an existing custom data model.

    Example
    -------
    >>> from fbnconfig.custom_data_model import CustomDataModelRef
    >>> model_ref = CustomDataModelRef(
    ...     id="my_ref",
    ...     entity_type="Instrument",
    ...     resource_id=ResourceId(scope="MyScope", code="MyModel")
    ... )
    """

    id: str = Field(exclude=True)
    entity_type: str
    resource_id: ResourceId

    def attach(self, client: httpx.Client):
        """Verify the custom data model exists."""
        try:
            (
                client.get(
                    url.format(
                        FULL_DATA_MODEL_URL,
                        entity_type=self.entity_type,
                        scope=self.resource_id.scope,
                        code=self.resource_id.code,
                    )
                ),
            )

        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(
                    f"CustomDataModel {self.entity_type}/{self.resource_id.scope}/"
                    f"{self.resource_id.code} not found"
                )
            else:
                raise ex


@register_resource()
class CustomDataModelResource(CamelAlias, BaseModel, Resource):
    """Resource describing a LUSID Custom Data Model.

    Creates a custom data model if it doesn't exist and updates if it has changed.
    Custom data models define validation rules, required properties, and identifiers
    for LUSID entities.

    Example
    -------
    >>> from fbnconfig import custom_data_model, Deployment
    >>> from fbnconfig.coretypes import ResourceId
    >>> model = custom_data_model.CustomDataModelResource(
    ...     id="my_model",
    ...     entity_type="Instrument",
    ...     resource_id=ResourceId(scope="MyScope", code="MyModel"),
    ...     display_name="My Custom Model",
    ...     description="A custom data model for instruments",
    ...     conditions="InstrumentDefinition.InstrumentType eq 'bond'",
    ...     properties=[
    ...         custom_data_model.DataModelProperty(
    ...             property_key="Instrument/MyScope/Rating",
    ...             required=True
    ...         )
    ...     ]
    ... )
    >>> Deployment("my_deployment", [model])

    Attributes
    ----------
    id : str
        Resource identifier; used in the log to reference the custom data model resource
    entity_type : str
        The entity type this model applies to (e.g., "Instrument", "Portfolio")
    resource_id : ResourceId
        The resource identifier containing scope and code
    display_name : str
        The display name for the custom data model
    description : str
        A description of the custom data model
    parent_data_model : ParentDataModelKey, optional
        Reference to a parent data model to inherit from
    conditions : str, optional
        Filter conditions for when this model applies
    properties : List[DataModelProperty], optional
        List of properties required or validated by this model
    identifier_types : List[IdentifierType], optional
        List of identifier types required or validated by this model
    attribute_aliases : List[AttributeAlias], optional
        List of attribute alias mappings
    recommended_sort_by : List[RecommendedSortBy], optional
        Recommended sorting configuration
    """

    id: str = Field(exclude=True)
    entity_type: str = Field(exclude=True)  # Path parameter, not in body
    resource_id: ResourceId = Field(serialization_alias="id")
    display_name: str
    description: str
    parent_data_model: ParentDataModelKey = None
    conditions: str | None = None
    properties: List[DataModelProperty] | None = None
    identifier_types: List[IdentifierType] | None = None
    attribute_aliases: List[AttributeAlias] | None = None
    recommended_sort_by: List[RecommendedSortBy] | None = None

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data: Any, info) -> Dict[str, Any]:
        if not isinstance(data, dict):
            return data

        # Handle nested id structure from API response first (for resource_id field)
        has_resource_id_in_data = "id" in data and isinstance(data["id"], dict) and "scope" in data["id"]

        if has_resource_id_in_data:
            # This is the resource_id coming from API
            data["resource_id"] = data.pop("id")

        # Handle id and entity_type from context (for dump/undump)
        # The context "id" is the resource identifier, not the resource_id
        if info.context:
            if info.context.get("id") and "id" not in data:
                # Only set if not already consumed from API data
                data["id"] = info.context.get("id")
            if info.context.get("entity_type"):
                data["entity_type"] = info.context.get("entity_type")

        return data

    def __get_content_hash__(self) -> str:
        dump = self.model_dump(
            mode="json", exclude_none=True, by_alias=True, exclude={"entity_type", "resource_id"}
        )
        return sha256(json.dumps(dump, sort_keys=True).encode()).hexdigest()

    def read(self, client: httpx.Client, old_state) -> Dict[str, Any]:
        return client.get(
            url.format(
                FULL_DATA_MODEL_URL,
                entity_type=old_state.entity_type,
                scope=old_state.scope,
                code=old_state.code,
            )
        ).json()

    def create(self, client: httpx.Client):
        desired = self.model_dump(
            mode="json",
            exclude_none=True,
            by_alias=True,
            exclude={"entity_type"},  # entity_type is path param
        )

        remote = client.post(
            url.format(DATA_MODEL_URL_FORMAT, entity_type=self.entity_type), json=desired
        ).json()

        return {
            "id": self.id,
            "entity_type": self.entity_type,
            "scope": self.resource_id.scope,
            "code": self.resource_id.code,
            "source_version": self.__get_content_hash__(),
            "remote_version": f"{remote['version']['asAtVersionNumber']}",
        }

    def update(self, client: httpx.Client, old_state):
        # Check if entity_type/scope/code changed (not allowed)
        if (
            self.entity_type != old_state.entity_type
            or self.resource_id.scope != old_state.scope
            or self.resource_id.code != old_state.code
        ):
            # Delete old and create new
            self.delete(client, old_state)
            return self.create(client)

        remote = self.read(client, old_state) or {}
        remote_version = f"{remote['version']['asAtVersionNumber']}"

        # Build request body (without id/resourceId for update)
        desired = self.model_dump(
            mode="json", exclude_none=True, by_alias=True, exclude={"entity_type", "resource_id"}
        )

        source_version = self.__get_content_hash__()

        # Check if content has changed using both source and remote versions
        if source_version == old_state.source_version and remote_version == old_state.remote_version:
            return None  # No change

        # Update the custom data model
        updated = client.put(
            url.format(
                FULL_DATA_MODEL_URL,
                entity_type=self.entity_type,
                scope=self.resource_id.scope,
                code=self.resource_id.code,
            ),
            json=desired,
        ).json()

        return {
            "id": self.id,
            "entity_type": self.entity_type,
            "scope": self.resource_id.scope,
            "code": self.resource_id.code,
            "source_version": source_version,
            "remote_version": f"{updated['version']['asAtVersionNumber']}",
        }

    @staticmethod
    def delete(client: httpx.Client, old_state):
        client.delete(
            url.format(
                FULL_DATA_MODEL_URL,
                entity_type=old_state.entity_type,
                scope=old_state.scope,
                code=old_state.code,
            )
        )

    def deps(self):
        """Return dependencies."""
        deps = []
        if self.parent_data_model and isinstance(self.parent_data_model, (Resource, Ref)):
            deps.append(self.parent_data_model)
        # Add property_key dependencies from properties
        if self.properties:
            for prop_def in self.properties:
                if isinstance(prop_def.property_key, (Resource, Ref)):
                    deps.append(prop_def.property_key)
        return deps



import pathlib
import textwrap
from unittest import mock

import httpx
import pytest
from click.testing import CliRunner
from starlette.testclient import TestClient

from server.app import app


@pytest.mark.respx(base_url="https://foo.lusid.com")
class DescribeServer:

    @pytest.fixture()
    def api(self):
        client = TestClient(app)
        client.event_hooks = {"response": [lambda x: x.raise_for_status()]}
        return client

    def test_homepage(self, api):
        response = api.get("/api/fbnconfig/")
        assert response.status_code == 200
        assert response.json() == {"message": "Hello, Banana!", "status": "success"}

    def test_log(self, api, respx_mock):
        # given lusid has no deployments
        respx_mock.get("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(200, json={
                "values": []
            })
        )
        # when we request them
        headers = {
            "Authorization": "Bearer BINGOBANGO",
            "X-LUSID-Host": "https://foo.lusid.com",
        }
        response = api.get("/api/fbnconfig/log/something", headers=headers)
        assert response.status_code == 200
        request = respx_mock.calls.last.request
        assert request.method == "GET"
        assert request.headers["Authorization"] == "Bearer BINGOBANGO"
        assert request.headers["Host"] == "foo.lusid.com"

    def test_schema(self, api):
        # when we get the schema
        s = api.get("/api/fbnconfig/schema").json()
        # then it has defs and properties at the top level as expected
        assert s
        assert "$defs" in s
        assert "properties" in s

    @pytest.fixture()
    def example_dir(self):
        runner = CliRunner()
        with runner.isolated_filesystem():
            # given there is one example script in the examples folder
            script = textwrap.dedent("""\
                from fbnconfig import Deployment
                def configure(env):
                    return Deployment('test1', [])
            """)
            example_dir = pathlib.Path.cwd()
            with open("script.py", "w") as f:
                f.write(script)
            yield example_dir

    def test_examples(self, api, example_dir):
        # given a dir with one example script
        # and we mock the server to use it
        with mock.patch("server.app.get_examples_path", return_value=example_dir):
            # when we request examples
            examples = api.get("/api/fbnconfig/examples/").json()
            # there is one example
            assert examples == [{
                "name": "script",
                "path": "/api/fbnconfig/examples/script"
            }]

    def test_one_example(self, api, example_dir):
        # given a dir with one example script
        # and we mock the server to use it
        with mock.patch("server.app.get_examples_path", return_value=example_dir):
            # when we request an example that exists
            example = api.get("/api/fbnconfig/examples/script").json()
            # then we get a json with zero resources
            assert example == {
                "deploymentId": "test1",
                "resources": []
            }



import json
from types import SimpleNamespace

import httpx
import pytest

from fbnconfig import property, side_definition, transaction_type

TEST_BASE = "https://foo.lusid.com"


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeTransactionType:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [lambda x: x.raise_for_status()]})

    @pytest.fixture
    def simple_transaction_type(self):
        return transaction_type.TransactionTypeResource(
            id="txn1",
            scope="sc1",
            source="default",
            aliases=[
                transaction_type.TransactionTypeAlias(
                    type="Buy",
                    description="Something",
                    transaction_class="default",
                    transaction_roles="LongLonger",
                    is_default=False,
                )
            ],
            movements=[
                transaction_type.TransactionTypeMovement(
                    movement_types=transaction_type.MovementType.StockMovement,
                    side=side_definition.SideResource(
                        id="side1",
                        side="Side1",
                        scope="sc1",
                        security="Txn:LusidInstrumentId",
                        currency="Txn:SettlementCurrency",
                        rate="Txn:TradeToPortfolioRate",
                        units="Txn:Units",
                        amount="Txn:TotalConsideration",
                        notional_amount="0",
                    ),
                    direction=1,
                    name="Stock Movement",
                    movement_options=[transaction_type.MovementOption.DirectAdjustment],
                    settlement_mode=transaction_type.SettlementMode.Internal,
                )
            ],
            calculations=[],
            properties=[],
        )

    @pytest.fixture
    def simple_transaction_type_no_movement_option(self):
        return transaction_type.TransactionTypeResource(
            id="txn1",
            scope="sc1",
            source="default",
            aliases=[
                transaction_type.TransactionTypeAlias(
                    type="Buy",
                    description="Something",
                    transaction_class="default",
                    transaction_roles="LongLonger",
                    is_default=False,
                )
            ],
            movements=[
                transaction_type.TransactionTypeMovement(
                    movement_types=transaction_type.MovementType.StockMovement,
                    side=side_definition.SideResource(
                        id="side1",
                        side="Side1",
                        scope="sc1",
                        security="Txn:LusidInstrumentId",
                        currency="Txn:SettlementCurrency",
                        rate="Txn:TradeToPortfolioRate",
                        units="Txn:Units",
                        amount="Txn:TotalConsideration",
                        notional_amount="0",
                    ),
                    direction=1,
                    name="Stock Movement",
                    settlement_mode=transaction_type.SettlementMode.Internal,
                )
            ],
            calculations=[],
            properties=[],
        )

    @pytest.fixture
    def complex_transaction_type(self):
        transaction_configuration_properties = [
            transaction_type.PerpetualProperty(
                property_key=property.DefinitionRef(
                    id="one", domain=property.Domain.TransactionConfiguration, scope="sc1", code="cd4"
                ),
                label_value="Hello world",
            ),
            transaction_type.PerpetualProperty(
                property_key=property.DefinitionRef(
                    id="one", domain=property.Domain.TransactionConfiguration, scope="sc1", code="cd5"
                ),
                metric_value=transaction_type.MetricValue(value=1.23456, unit="GBP"),
            ),
            transaction_type.PerpetualProperty(
                property_key=property.DefinitionRef(
                    id="one", domain=property.Domain.TransactionConfiguration, scope="sc1", code="cd6"
                ),
                label_set_value=["one", "two", "three"],
            ),
        ]
        return transaction_type.TransactionTypeResource(
            aliases=[
                transaction_type.TransactionTypeAlias(
                    type="Buy",
                    description="Test buy 1",
                    transaction_class="ConfigurationTest",
                    transaction_roles="AllRoles",
                    is_default=False,
                ),
                transaction_type.TransactionTypeAlias(
                    type="BY",
                    description="Test buy 2",
                    transaction_class="ConfigurationTest",
                    transaction_roles="AllRoles",
                    is_default=False,
                ),
            ],
            movements=[
                transaction_type.TransactionTypeMovement(
                    movement_types=transaction_type.MovementType.StockMovement,
                    side=side_definition.SideRef(id="side1", side="Side1", scope="sc1"),
                    direction=1,
                    name="Stock Movement",
                    movement_options=[transaction_type.MovementOption.DirectAdjustment],
                    mappings=[
                        transaction_type.TransactionTypePropertyMapping(
                            property_key=property.DefinitionRef(
                                id="one", domain=property.Domain.Transaction, scope="sc1", code="cd1"
                            ),
                            set_to="Hello world",
                        ),
                        transaction_type.TransactionTypePropertyMapping(
                            property_key=property.DefinitionRef(
                                id="two", domain=property.Domain.Transaction, scope="sc1", code="cd2"
                            ),
                            map_from=property.DefinitionRef(
                                id="three", domain=property.Domain.Transaction, scope="sc1", code="cd3"
                            ),
                        ),
                    ],
                    properties=transaction_configuration_properties,
                    settlement_mode=transaction_type.SettlementMode.Internal,
                )
            ],
            calculations=[
                transaction_type.TransactionTypeCalculation(
                    type=transaction_type.CalculationType.TaxAmounts,
                    side=side_definition.SideRef(id="side1", side="Side1", scope="sc1"),
                ),
                transaction_type.TransactionTypeCalculation(
                    type=transaction_type.CalculationType.NotionalAmount
                ),
            ],
            properties=transaction_configuration_properties,
            id="txn1",
            scope="sc1",
            source="default",
        )

    def test_read_transaction_type(self, respx_mock, simple_transaction_type):
        respx_mock.get(
            "/api/api/transactionconfiguration/types/default/Buy"
        ).mock(
            return_value=httpx.Response(
                200,
                json={"aliases": [], "movements": [], "links": [{"x": "www.x.com"}]}
            )
        )
        # given an existing transaction type
        old_state = SimpleNamespace(transaction_type="Buy", source="default", scope="sc1")
        sut = simple_transaction_type
        # when it reads
        response = sut.read(self.client, old_state)
        # then it builds the correct url with the scope in params
        assert respx_mock.calls.last.request.url == "https://foo.lusid.com/api/api/transactionconfiguration/types/default/Buy?scope=sc1"
        # and it removes the links from the response
        assert response == {
            "aliases": [
            ],
            "movements": [
            ],
        }

    def test_read_transaction_type_missing(self, respx_mock, simple_transaction_type):
        # given a transaction type that does not exist at the server
        respx_mock.get(
            f"/api/api/transactionconfiguration/types/{simple_transaction_type.source}/{simple_transaction_type._get_first_alias()}?scope={simple_transaction_type.scope}"
        ).mock(return_value=httpx.Response(404))
        # and we construct a desired state
        old_state = SimpleNamespace(transaction_type="Buy", source="default", scope="sc1")
        sut = simple_transaction_type
        # when we try to read it, it throws
        with pytest.raises(httpx.HTTPError):
            sut.read(self.client, old_state)

    def test_create_transaction_type(self, respx_mock, simple_transaction_type):
        respx_mock.put(
            f"/api/api/transactionconfiguration/types/{simple_transaction_type.source}/{simple_transaction_type._get_first_alias()}?scope={simple_transaction_type.scope}"
        ).mock(
            return_value=httpx.Response(
                200, json={"source": "default", "scope": "sc1", "transaction_type": "Buy"}
            )
        )

        client = self.client
        sut = simple_transaction_type

        state = sut.create(client)
        req = respx_mock.calls.last.request

        assert json.loads(req.content) == {
            "aliases": [
                {
                    "type": "Buy",
                    "description": "Something",
                    "transactionClass": "default",
                    "transactionRoles": "LongLonger",
                    "isDefault": False,
                }
            ],
            "movements": [
                {
                    "movementTypes": "StockMovement",
                    "side": "Side1",
                    "direction": 1,
                    "properties": {},
                    "mappings": [],
                    "name": "Stock Movement",
                    "movementOptions": ["DirectAdjustment"],
                    "condition": "",
                    "settlementMode": "Internal",
                }
            ],
            "properties": {},
            "calculations": [],
        }
        assert state == {"source": "default", "scope": "sc1", "transaction_type": "Buy"}

    def test_complex_transaction_type_create(self, respx_mock, complex_transaction_type):
        respx_mock.put(
            "/api/api/transactionconfiguration/types/default/BY?scope=sc1"
        ).mock(
            return_value=httpx.Response(
                200, json={"source": "default", "scope": "sc1", "transaction_type": "Buy"}
            )
        )
        # given a complex transaction type
        sut = complex_transaction_type
        # when we create it
        state = sut.create(self.client)
        # then the state is unchanged
        assert state == {"source": "default", "scope": "sc1", "transaction_type": "BY"}
        # and the request body is correct
        req = respx_mock.calls.last.request
        assert json.loads(req.content) == {
            "aliases": [{
                "type": "Buy",
                "description": "Test buy 1",
                "transactionClass": "ConfigurationTest",
                "transactionRoles": "AllRoles",
                "isDefault": False,
            }, {
                "type": "BY",
                "description": "Test buy 2",
                "transactionClass": "ConfigurationTest",
                "transactionRoles": "AllRoles",
                "isDefault": False,
            }, ],
            "movements": [{
                "movementTypes": "StockMovement",
                "side": "Side1",
                "direction": 1,
                "settlementMode": "Internal",
                "properties": {
                    "TransactionConfiguration/sc1/cd4": {
                        "key": "TransactionConfiguration/sc1/cd4",
                        "value": {"labelValue": "Hello world"},
                    },
                    "TransactionConfiguration/sc1/cd5": {
                        "key": "TransactionConfiguration/sc1/cd5",
                        "value": {"metricValue": {"value": 1.23456, "unit": "GBP"}},
                    },
                    "TransactionConfiguration/sc1/cd6": {
                        "key": "TransactionConfiguration/sc1/cd6",
                        "value": {"labelSetValue": ["one", "two", "three"]},
                    },
                },
                "mappings": [
                    {"propertyKey": "Transaction/sc1/cd1", "setTo": "Hello world"},
                    {"propertyKey": "Transaction/sc1/cd2", "mapFrom": "Transaction/sc1/cd3"},
                ],
                "name": "Stock Movement",
                "movementOptions": ["DirectAdjustment"],
                "condition": "",
            }],
            "properties": {
                "TransactionConfiguration/sc1/cd4": {
                    "key": "TransactionConfiguration/sc1/cd4",
                    "value": {"labelValue": "Hello world"},
                },
                "TransactionConfiguration/sc1/cd5": {
                    "key": "TransactionConfiguration/sc1/cd5",
                    "value": {"metricValue": {"value": 1.23456, "unit": "GBP"}},
                },
                "TransactionConfiguration/sc1/cd6": {
                    "key": "TransactionConfiguration/sc1/cd6",
                    "value": {"labelSetValue": ["one", "two", "three"]},
                },
            },
            "calculations": [{"type": "TaxAmounts", "side": "Side1"}, {"type": "Txn:NotionalAmount"}],
        }

    def test_delete_transaction_type(self, respx_mock, simple_transaction_type):
        respx_mock.delete(
            f"/api/api/transactionconfiguration/types/{simple_transaction_type.source}/{simple_transaction_type._get_first_alias()}?scope={simple_transaction_type.scope}"
        ).mock(return_value=httpx.Response(200))
        client = self.client
        old_state = SimpleNamespace(transaction_type="Buy", source="default", scope="sc1")
        sut = simple_transaction_type

        sut.delete(client, old_state)

    def test_delete_transaction_type_missing(self, respx_mock, simple_transaction_type):
        respx_mock.delete(
            f"/api/api/transactionconfiguration/types/{simple_transaction_type.source}/{simple_transaction_type._get_first_alias()}?scope={simple_transaction_type.scope}"
        ).mock(return_value=httpx.Response(404))
        client = self.client
        with pytest.raises(httpx.HTTPError):
            old_state = SimpleNamespace(transaction_type="Buy", source="default", scope="sc1")
            sut = simple_transaction_type
            sut.delete(client, old_state)

    def test_update_transaction_type_no_change(self, respx_mock, simple_transaction_type):
        respx_mock.get(
            f"/api/api/transactionconfiguration/types/{simple_transaction_type.source}/{simple_transaction_type._get_first_alias()}?scope={simple_transaction_type.scope}"
        ).mock(
            return_value=httpx.Response(
                200,
                json={
                    "aliases": [
                        {
                            "type": "Buy",
                            "description": "Something",
                            "transactionClass": "default",
                            "transactionRoles": "LongLonger",
                            "isDefault": False,
                        }
                    ],
                    "movements": [
                        {
                            "movementTypes": "StockMovement",
                            "side": "Side1",
                            "direction": 1,
                            "properties": {},
                            "mappings": [],
                            "name": "Stock Movement",
                            "movementOptions": ["DirectAdjustment"],
                            "condition": "",
                            "settlementMode": "Internal",
                        }
                    ],
                    "properties": {},
                    "calculations": [],
                },
            )
        )

        client = self.client
        old_state = SimpleNamespace(transaction_type="Buy", source="default", scope="sc1")
        sut = simple_transaction_type

        state = sut.update(client, old_state)

        assert state is None

    def test_update_transaction_type_no_change_no_movement_option(
        self, respx_mock, simple_transaction_type_no_movement_option
    ):
        respx_mock.get(
            f"/api/api/transactionconfiguration/types/{simple_transaction_type_no_movement_option.source}/"
            f"{simple_transaction_type_no_movement_option._get_first_alias()}"
            f"?scope={simple_transaction_type_no_movement_option.scope}"
        ).mock(
            return_value=httpx.Response(
                200,
                json={
                    "aliases": [
                        {
                            "type": "Buy",
                            "description": "Something",
                            "transactionClass": "default",
                            "transactionRoles": "LongLonger",
                            "isDefault": False,
                        }
                    ],
                    "movements": [
                        {
                            "movementTypes": "StockMovement",
                            "side": "Side1",
                            "direction": 1,
                            "properties": {},
                            "mappings": [],
                            "name": "Stock Movement",
                            "movementOptions": [],
                            "condition": "",
                            "settlementMode": "Internal",
                        }
                    ],
                    "properties": {},
                    "calculations": [],
                },
            )
        )

        client = self.client
        old_state = SimpleNamespace(transaction_type="Buy", source="default", scope="sc1")
        sut = simple_transaction_type_no_movement_option

        state = sut.update(client, old_state)

        assert state is None

    def test_update_transaction_type_state_change(self, respx_mock, simple_transaction_type):
        respx_mock.delete(
            f"/api/api/transactionconfiguration/types/{simple_transaction_type.source}/{simple_transaction_type._get_first_alias()}?scope=sc2"
        ).mock(return_value=httpx.Response(200))

        respx_mock.put(
            f"/api/api/transactionconfiguration/types/{simple_transaction_type.source}/{simple_transaction_type._get_first_alias()}?scope={simple_transaction_type.scope}"
        ).mock(
            return_value=httpx.Response(
                200, json={"source": "default", "scope": "sc1", "transaction_type": "Buy"}
            )
        )

        client = self.client

        old_state = SimpleNamespace(transaction_type="Buy", source="default", scope="sc2")
        sut = simple_transaction_type
        state = sut.update(client, old_state)
        assert state == {"source": "default", "scope": "sc1", "transaction_type": "Buy"}

    def test_update_transaction_type_content_changed(self, respx_mock, simple_transaction_type):
        respx_mock.get(
            f"/api/api/transactionconfiguration/types/{simple_transaction_type.source}/{simple_transaction_type._get_first_alias()}?scope={simple_transaction_type.scope}"
        ).mock(
            return_value=httpx.Response(
                200,
                json=simple_transaction_type.model_dump(mode="json", exclude_none=True, by_alias=True),
            )
        )

        respx_mock.put(
            f"/api/api/transactionconfiguration/types/{simple_transaction_type.source}/{simple_transaction_type._get_first_alias()}?scope={simple_transaction_type.scope}"
        ).mock(
            return_value=httpx.Response(
                200, json={"source": "default", "scope": "sc1", "transaction_type": "Buy"}
            )
        )

        client = self.client

        old_state = SimpleNamespace(transaction_type="Buy", source="default", scope="sc1")

        sut = simple_transaction_type

        sut.movements = []
        state = sut.update(client, old_state)
        assert state == {"source": "default", "scope": "sc1", "transaction_type": "Buy"}

    def test_deps_movements_unique(self):
        side_resource = side_definition.SideResource(
            id="side1",
            side="Side1",
            scope="sc1",
            security="Txn:LusidInstrumentId",
            currency="Txn:SettlementCurrency",
            rate="Txn:TradeToPortfolioRate",
            units="Txn:Units",
            amount="Txn:TotalConsideration",
            notional_amount="0",
        )

        side_ref = side_definition.SideRef(id="side1", side="Side1", scope="sc1")

        txn_type_no_movements = transaction_type.TransactionTypeResource(
            id="txn1",
            scope="sc1",
            source="default",
            aliases=[
                transaction_type.TransactionTypeAlias(
                    type="Buy",
                    description="Something",
                    transaction_class="default",
                    transaction_roles="LongLonger",
                    is_default=False,
                )
            ],
            movements=[],
            calculations=[],
            properties=[],
        )

        txn_type_side_resource_movement = transaction_type.TransactionTypeResource(
            id="txn1",
            scope="sc1",
            source="default",
            aliases=[
                transaction_type.TransactionTypeAlias(
                    type="Buy",
                    description="Something",
                    transaction_class="default",
                    transaction_roles="LongLonger",
                    is_default=False,
                )
            ],
            movements=[
                transaction_type.TransactionTypeMovement(
                    movement_types=transaction_type.MovementType.StockMovement,
                    side=side_resource,
                    direction=1,
                    name="Stock Movement",
                    movement_options=[transaction_type.MovementOption.DirectAdjustment],
                    settlement_mode=transaction_type.SettlementMode.Internal,
                ),
                transaction_type.TransactionTypeMovement(
                    movement_types=transaction_type.MovementType.StockMovement,
                    side=side_resource,
                    direction=1,
                    name="Stock Movement 2",
                    movement_options=[transaction_type.MovementOption.DirectAdjustment],
                    settlement_mode=transaction_type.SettlementMode.Internal,
                ),
            ],
            calculations=[],
            properties=[],
        )

        txn_type_side_ref_movement = transaction_type.TransactionTypeResource(
            id="txn1",
            scope="sc1",
            source="default",
            aliases=[
                transaction_type.TransactionTypeAlias(
                    type="Buy",
                    description="Something",
                    transaction_class="default",
                    transaction_roles="LongLonger",
                    is_default=False,
                )
            ],
            movements=[
                transaction_type.TransactionTypeMovement(
                    movement_types=transaction_type.MovementType.StockMovement,
                    side=side_ref,
                    direction=1,
                    name="Stock Movement",
                    movement_options=[transaction_type.MovementOption.DirectAdjustment],
                    settlement_mode=transaction_type.SettlementMode.Internal,
                ),
                transaction_type.TransactionTypeMovement(
                    movement_types=transaction_type.MovementType.StockMovement,
                    side=side_ref,
                    direction=1,
                    name="Stock Movement 2",
                    movement_options=[transaction_type.MovementOption.DirectAdjustment],
                    settlement_mode=transaction_type.SettlementMode.Internal,
                ),
            ],
            calculations=[],
            properties=[],
        )

        deps_no_movements = txn_type_no_movements.deps()
        deps_side_resource_movement = txn_type_side_resource_movement.deps()
        deps_side_ref_movement = txn_type_side_ref_movement.deps()

        assert deps_no_movements == []
        assert deps_side_resource_movement == [side_resource]
        assert deps_side_ref_movement == [side_ref]

    def test_deps_movement_properties_unique(self):
        transaction_configuration_properties = [
            transaction_type.PerpetualProperty(
                property_key=property.DefinitionResource(
                    id="pd1",
                    domain=property.Domain.TransactionConfiguration,
                    scope="sc1",
                    code="pd1",
                    display_name="property",
                    data_type_id=property.ResourceId(scope="system", code="number"),
                    constraint_style=property.ConstraintStyle.Property,
                    property_description="property",
                    life_time=property.LifeTime.Perpetual,
                    collection_type=None,
                ),
                label_value="Hello world",
            ),
            transaction_type.PerpetualProperty(
                property_key=property.DefinitionRef(
                    id="one", domain=property.Domain.TransactionConfiguration, scope="sc1", code="cd5"
                ),
                metric_value=transaction_type.MetricValue(value=1.23456, unit="GBP"),
            ),
            transaction_type.PerpetualProperty(
                property_key=property.DefinitionRef(
                    id="one", domain=property.Domain.TransactionConfiguration, scope="sc1", code="cd5"
                ),
                label_set_value=["one", "two", "three"],
            ),
        ]

        side_ref = side_definition.SideRef(id="side1", side="Side1", scope="sc1")

        txn_type = transaction_type.TransactionTypeResource(
            id="txn1",
            scope="sc1",
            source="default",
            aliases=[
                transaction_type.TransactionTypeAlias(
                    type="Buy",
                    description="Something",
                    transaction_class="default",
                    transaction_roles="LongLonger",
                    is_default=False,
                )
            ],
            movements=[
                transaction_type.TransactionTypeMovement(
                    movement_types=transaction_type.MovementType.StockMovement,
                    side=side_ref,
                    direction=1,
                    name="Stock Movement",
                    movement_options=[transaction_type.MovementOption.DirectAdjustment],
                    properties=None,
                    settlement_mode=transaction_type.SettlementMode.Internal,
                )
            ],
            calculations=[],
            properties=[],
        )

        assert txn_type.deps() == [side_ref]

        txn_type.movements[0].properties = transaction_configuration_properties

        assert txn_type.deps() == [
            side_ref,
            property.DefinitionResource(
                id="pd1",
                domain=property.Domain.TransactionConfiguration,
                scope="sc1",
                code="pd1",
                display_name="property",
                data_type_id=property.ResourceId(scope="system", code="number"),
                constraint_style=property.ConstraintStyle.Property,
                property_description="property",
                life_time=property.LifeTime.Perpetual,
                collection_type=None,
            ),
            property.DefinitionRef(
                id="one", domain=property.Domain.TransactionConfiguration, scope="sc1", code="cd5"
            ),
        ]

    def test_deps_movements_mappings_unique(self):
        side_ref = side_definition.SideRef(id="side1", side="Side1", scope="sc1")

        mappings = [
            transaction_type.TransactionTypePropertyMapping(
                property_key=property.DefinitionRef(
                    id="one", domain=property.Domain.Transaction, scope="sc1", code="cd1"
                ),
                set_to="Hello world",
            ),
            transaction_type.TransactionTypePropertyMapping(
                property_key=property.DefinitionRef(
                    id="two", domain=property.Domain.Transaction, scope="sc1", code="cd2"
                ),
                map_from=property.DefinitionRef(
                    id="three", domain=property.Domain.Transaction, scope="sc1", code="cd3"
                ),
            ),
        ]

        txn_type = transaction_type.TransactionTypeResource(
            id="txn1",
            scope="sc1",
            source="default",
            aliases=[
                transaction_type.TransactionTypeAlias(
                    type="Buy",
                    description="Something",
                    transaction_class="default",
                    transaction_roles="LongLonger",
                    is_default=False,
                )
            ],
            movements=[
                transaction_type.TransactionTypeMovement(
                    movement_types=transaction_type.MovementType.StockMovement,
                    side=side_ref,
                    direction=1,
                    name="Stock Movement",
                    movement_options=[transaction_type.MovementOption.DirectAdjustment],
                    settlement_mode=transaction_type.SettlementMode.Internal,
                )
            ],
            calculations=[],
            properties=[],
        )

        assert txn_type.deps() == [side_ref]

        txn_type.movements[0].mappings = mappings

        assert txn_type.deps() == [
            side_ref,
            property.DefinitionRef(
                id="one", domain=property.Domain.Transaction, scope="sc1", code="cd1"
            ),
            property.DefinitionRef(
                id="two", domain=property.Domain.Transaction, scope="sc1", code="cd2"
            ),
            property.DefinitionRef(
                id="three", domain=property.Domain.Transaction, scope="sc1", code="cd3"
            ),
        ]

    def test_deps_properties_unique(self):
        side_ref = side_definition.SideRef(id="side1", side="Side1", scope="sc1")

        properties = [
            transaction_type.PerpetualProperty(
                property_key=property.DefinitionRef(
                    id="one", domain=property.Domain.TransactionConfiguration, scope="sc1", code="cd1"
                ),
                label_value="Hello world",
            ),
            transaction_type.PerpetualProperty(
                property_key=property.DefinitionRef(
                    id="two", domain=property.Domain.TransactionConfiguration, scope="sc1", code="cd2"
                ),
                metric_value=transaction_type.MetricValue(value=1.23456, unit="GBP"),
            ),
            transaction_type.PerpetualProperty(
                property_key=property.DefinitionRef(
                    id="two", domain=property.Domain.TransactionConfiguration, scope="sc1", code="cd2"
                ),
                label_set_value=["one", "two", "three"],
            ),
        ]

        txn_type = transaction_type.TransactionTypeResource(
            id="txn1",
            scope="sc1",
            source="default",
            aliases=[
                transaction_type.TransactionTypeAlias(
                    type="Buy",
                    description="Something",
                    transaction_class="default",
                    transaction_roles="LongLonger",
                    is_default=False,
                )
            ],
            movements=[
                transaction_type.TransactionTypeMovement(
                    movement_types=transaction_type.MovementType.StockMovement,
                    side=side_ref,
                    direction=1,
                    name="Stock Movement",
                    movement_options=[transaction_type.MovementOption.DirectAdjustment],
                    settlement_mode=transaction_type.SettlementMode.Internal,
                )
            ],
            calculations=[],
            properties=[],
        )

        assert txn_type.deps() == [side_ref]

        txn_type.properties = properties

        assert txn_type.deps() == [
            side_ref,
            property.DefinitionRef(
                id="one", domain=property.Domain.TransactionConfiguration, scope="sc1", code="cd1"
            ),
            property.DefinitionRef(
                id="two", domain=property.Domain.TransactionConfiguration, scope="sc1", code="cd2"
            ),
        ]

    def test_deps_calculations_unique(self):
        side_ref = side_definition.SideRef(id="side1", side="Side1", scope="sc1")

        side_resource = side_definition.SideResource(
            id="side2",
            side="Side2",
            scope="sc2",
            security="Txn:LusidInstrumentId",
            currency="Txn:SettlementCurrency",
            rate="Txn:TradeToPortfolioRate",
            units="Txn:Units",
            amount="Txn:TotalConsideration",
            notional_amount="0",
        )

        calculations = [
            transaction_type.TransactionTypeCalculation(
                type=transaction_type.CalculationType.TaxAmounts, side=side_ref
            ),
            transaction_type.TransactionTypeCalculation(
                type=transaction_type.CalculationType.ExchangeRate, side=side_ref
            ),
            transaction_type.TransactionTypeCalculation(
                type=transaction_type.CalculationType.TaxAmounts, side=side_resource
            ),
            transaction_type.TransactionTypeCalculation(
                type=transaction_type.CalculationType.NotionalAmount
            ),
        ]

        txn_type = transaction_type.TransactionTypeResource(
            id="txn1",
            scope="sc1",
            source="default",
            aliases=[
                transaction_type.TransactionTypeAlias(
                    type="Buy",
                    description="Something",
                    transaction_class="default",
                    transaction_roles="LongLonger",
                    is_default=False,
                )
            ],
            movements=[
                transaction_type.TransactionTypeMovement(
                    movement_types=transaction_type.MovementType.StockMovement,
                    side=side_ref,
                    direction=1,
                    name="Stock Movement",
                    movement_options=[transaction_type.MovementOption.DirectAdjustment],
                    settlement_mode=transaction_type.SettlementMode.Internal,
                )
            ],
            calculations=[],
            properties=[],
        )

        assert txn_type.deps() == [side_ref]

        txn_type.calculations = calculations

        assert txn_type.deps() == [side_ref, side_resource]

    def test_transaction_type_property_only_one_exists(self):
        # When we set more than one of label_value, label_set_value or metric_value
        # on a property, an error is thrown
        with pytest.raises(KeyError):
            transaction_type.PerpetualProperty(
                property_key=property.DefinitionRef(
                    id="one", domain=property.Domain.Transaction, scope="sc1", code="cd1"
                ),
                label_value="Hello world",
                metric_value=transaction_type.MetricValue(value=1.23456, unit="GBP"),
            )

        # If three exist, throw error
        with pytest.raises(KeyError):
            transaction_type.PerpetualProperty(
                property_key=property.DefinitionRef(
                    id="one", domain=property.Domain.Transaction, scope="sc1", code="cd1"
                ),
                label_value="Hello world",
                label_set_value=["Hello", "World"],
                metric_value=transaction_type.MetricValue(value=1.23456, unit="GBP"),
            )

        # When one exists, it builds without error
        transaction_type.PerpetualProperty(
            property_key=property.DefinitionRef(
                id="one", domain=property.Domain.Transaction, scope="sc1", code="cd1"
            ),
            metric_value=transaction_type.MetricValue(value=1.23456, unit="GBP"),
        )

    def test_no_aliases_throws_value_error(self):
        with pytest.raises(ValueError):
            transaction_type.TransactionTypeResource(
                id="txn1", scope="sc1", source="default", aliases=[]
            )

    def test_equal_json_with_different_ordering(self):
        json1 = {
            "aliases": [
                {
                    "type": "ZZZZBuy",
                    "description": "Test buy 1",
                    "transactionClass": "ConfigurationTest",
                    "transactionRoles": "AllRoles",
                    "isDefault": False,
                },
                {
                    "type": "BY",
                    "description": "Test buy 2",
                    "transactionClass": "ConfigurationTest",
                    "transactionRoles": "AllRoles",
                    "isDefault": False,
                },
            ],
            "calculations": [
                {"type": "TaxAmounts", "side": "Side1"},
                {"type": "AAAAATaxAmounts", "side": "Side1"},
            ],
        }

        json2 = {
            "aliases": [
                {
                    "type": "BY",
                    "description": "Test buy 2",
                    "transactionClass": "ConfigurationTest",
                    "transactionRoles": "AllRoles",
                    "isDefault": False,
                },
                {
                    "type": "ZZZZBuy",
                    "description": "Test buy 1",
                    "transactionClass": "ConfigurationTest",
                    "transactionRoles": "AllRoles",
                    "isDefault": False,
                },
            ],
            "calculations": [
                {"type": "AAAAATaxAmounts", "side": "Side1"},
                {"type": "TaxAmounts", "side": "Side1"},
            ],
        }

        assert transaction_type._compare_json_structures(json1, json2)

    def test_not_equal_json(self):
        json1 = {
            "aliases": [
                {
                    "type": "ZZZZBuy",
                    "description": "Test buy 1",
                    "transactionClass": "ConfigurationTest",
                    "transactionRoles": "AllRoles",
                    "isDefault": False,
                },
                {
                    "type": "BY",
                    "description": "Test buy 2",
                    "transactionClass": "ConfigurationTest",
                    "transactionRoles": "AllRoles",
                    "isDefault": False,
                },
            ]
        }

        json2 = {
            "aliases": [
                {
                    "type": "BY",
                    "description": "Test buy 2",
                    "transactionClass": "ConfigurationTest",
                    "transactionRoles": "AllRoles",
                    "isDefault": False,
                },
                {
                    "type": "ZZZZBuy",
                    "description": "Test buy 1",
                    "transactionClass": "ConfigurationTest",
                    "transactionRoles": "AllRoles",
                    "isDefault": True,  # Difference here
                },
            ]
        }

        assert not transaction_type._compare_json_structures(json1, json2)

    def test_nested_json_with_different_ordering(self):
        json1 = {
            "movements": [
                {
                    "movementTypes": "StockMovement",
                    "side": "Side1",
                    "direction": 1,
                    "properties": {
                        "TransactionConfiguration/sc1/cd4": {
                            "key": "TransactionConfiguration/sc1/cd4",
                            "value": {"labelValue": "Hello world"},
                        },
                        "TransactionConfiguration/sc1/cd5": {
                            "key": "TransactionConfiguration/sc1/cd5",
                            "value": {"metricValue": {"value": 1.23456, "unit": "GBP"}},
                        },
                    },
                    "mappings": [
                        {"propertyKey": "Transaction/sc1/cd2", "mapFrom": "Transaction/sc1/cd3"},
                        {"propertyKey": "Transaction/sc1/cd1", "setTo": "Hello world"},
                    ],
                    "name": "Stock Movement",
                    "movementOptions": ["IncludeTaxLots", "DirectAdjustment"],
                    "condition": "",
                    "settlementMode": "Internal",
                }
            ]
        }

        json2 = {
            "movements": [
                {
                    "movementTypes": "StockMovement",
                    "side": "Side1",
                    "direction": 1,
                    "properties": {
                        "TransactionConfiguration/sc1/cd5": {
                            "key": "TransactionConfiguration/sc1/cd5",
                            "value": {"metricValue": {"value": 1.23456, "unit": "GBP"}},
                        },
                        "TransactionConfiguration/sc1/cd4": {
                            "key": "TransactionConfiguration/sc1/cd4",
                            "value": {"labelValue": "Hello world"},
                        },
                    },
                    "mappings": [
                        {"propertyKey": "Transaction/sc1/cd1", "setTo": "Hello world"},
                        {"propertyKey": "Transaction/sc1/cd2", "mapFrom": "Transaction/sc1/cd3"},
                    ],
                    "name": "Stock Movement",
                    "movementOptions": ["DirectAdjustment", "IncludeTaxLots"],
                    "condition": "",
                    "settlementMode": "Internal",
                }
            ]
        }

        assert transaction_type._compare_json_structures(json1, json2)

    def test_different_json_structure(self):
        json1 = {
            "movements": [
                {
                    "movementTypes": "StockMovement",
                    "side": "Side1",
                    "direction": 1,
                    "settlementMode": "Internal",
                }
            ]
        }
        json2 = {
            "movements": [
                {
                    "movementTypes": "StockMovement",
                    "side": "Side2",  # Different side
                    "direction": 1,
                    "settlementMode": "Internal",
                }
            ]
        }
        assert not transaction_type._compare_json_structures(json1, json2)

    def test_roundtrip_api_format(self):
        prop1 = property.DefinitionRef(
            id="prop1", domain=property.Domain.Transaction, scope="sc1", code="TestProp"
        )
        prop2 = property.DefinitionRef(
            id="prop2", domain=property.Domain.Transaction, scope="sc1", code="MapProp"
        )
        side1 = side_definition.SideRef(
            id="calc_side", side="CalcSide", scope="sc1"
        )
        # given API response format with scope and source included
        api_data = {
            "scope": "sc1",
            "source": "default",
            "aliases": [
                {
                    "type": "Buy",
                    "description": "Something",
                    "transaction_class": "default",
                    "transaction_roles": "LongLonger",
                    "is_default": False,
                }
            ],
            "movements": [
                {
                    "movement_types": "StockMovement",
                    "side": {"$ref": side1.id},
                    "direction": 1,
                    "properties": [
                        {
                            "property_key": {"$ref": "prop1"},
                            "label_value": "Test Label"
                        }
                    ],
                    "mappings": [
                        {
                            "property_key": {"$ref": "prop2"},
                            "set_to": "Test Value"
                        }
                    ],
                    "name": "Stock Movement",
                    "movement_options": ["DirectAdjustment"],
                    "condition": "",
                }
            ],
            "properties": [
                {
                    "property_key": {"$ref": "prop1"},
                    "metric_value": {"value": 123.45, "unit": "USD"}
                }
            ],
            "calculations": [
                {
                    "type": "TaxAmounts",
                    "side": {"$ref": side1.id},
                    "formula": "test_formula"
                }
            ],
        }
        # when we parse it back into a TransactionTypeResource with id in context
        result = transaction_type.TransactionTypeResource.model_validate(
            api_data, context={
                "id": "test-txn",
                "$refs": {
                    "prop1": prop1,
                    "prop2": prop2,
                    "calc_side": side1,
                }
            }
        )
        # then it's correctly populated
        assert result.id == "test-txn"
        assert result.scope == "sc1"
        assert result.source == "default"
        assert len(result.aliases) == 1
        assert result.aliases[0].type == "Buy"
        # verify movements with properties and mappings
        assert len(result.movements) == 1
        movement = result.movements[0]
        assert movement.side.side == "CalcSide"
        assert movement.properties
        assert len(movement.properties) == 1
        assert movement.properties[0].property_key.code == "TestProp"
        assert movement.properties[0].label_value == "Test Label"
        assert movement.mappings
        assert len(movement.mappings) == 1
        assert movement.mappings[0].property_key.code == "MapProp"
        assert movement.mappings[0].set_to == "Test Value"
        # verify global properties
        assert result.properties
        assert len(result.properties) == 1
        assert result.properties[0].property_key.code == "TestProp"
        assert result.properties[0].metric_value
        assert result.properties[0].metric_value.value == 123.45
        assert result.properties[0].metric_value.unit == "USD"
        assert result.calculations
        # verify calculations with side reference
        assert len(result.calculations) == 1
        assert result.calculations[0].type == "TaxAmounts"
        assert result.calculations[0].side
        assert result.calculations[0].side.side == "CalcSide"
        assert result.calculations[0].formula == "test_formula"

    def test_dump(self, simple_transaction_type):
        # given a transaction type resource
        sut = simple_transaction_type
        # when we dump it
        result = sut.model_dump(
            mode="json",
            by_alias=True,
            round_trip=True,
            exclude_none=True,
            context={"style": "dump"}
        )
        # then it includes scope and source information
        expected_keys = {"scope", "source", "aliases", "movements", "properties", "calculations"}
        assert set(result.keys()) == expected_keys
        assert result["scope"] == "sc1"
        assert result["source"] == "default"
        assert len(result["aliases"]) == 1
        assert result["aliases"][0]["type"] == "Buy"

    def test_undump(self):
        # given dump data
        data = {
            "scope": "sc1",
            "source": "default",
            "aliases": [
                {
                    "type": "Buy",
                    "description": "Something",
                    "transaction_class": "default",
                    "transaction_roles": "LongLonger",
                    "is_default": False,
                }
            ],
            "movements": [],
            "properties": [],
            "calculations": [],
        }
        # when we undump it
        result = transaction_type.TransactionTypeResource.model_validate(
            data, context={"style": "undump", "id": "txn_id"}
        )
        # then it's correctly populated
        assert result.id == "txn_id"
        assert result.scope == "sc1"
        assert result.source == "default"
        assert len(result.aliases) == 1
        assert result.aliases[0].type == "Buy"

    def test_dump_with_refs(self):
        # given a transaction type resource with property and side references
        prop1 = property.DefinitionRef(
            id="prop1", domain=property.Domain.Transaction, scope="sc1", code="TestProp"
        )
        side1 = side_definition.SideRef(
            id="side1", side="TestSide", scope="sc1"
        )
        sut = transaction_type.TransactionTypeResource(
            id="txn_with_refs",
            scope="sc1",
            source="default",
            aliases=[
                transaction_type.TransactionTypeAlias(
                    type="Buy",
                    description="Test transaction",
                    transaction_class="default",
                    transaction_roles="LongLonger",
                    is_default=False,
                )
            ],
            movements=[
                transaction_type.TransactionTypeMovement(
                    movement_types=transaction_type.MovementType.StockMovement,
                    side=side1,
                    direction=1,
                    properties=[
                        transaction_type.PerpetualProperty(
                            property_key=prop1,
                            label_value="Test Label"
                        )
                    ],
                    mappings=[
                        transaction_type.TransactionTypePropertyMapping(
                            property_key=prop1,
                            set_to="Test Value"
                        )
                    ],
                    name="Stock Movement",
                    movement_options=[transaction_type.MovementOption.DirectAdjustment],
                )
            ],
            properties=[
                transaction_type.PerpetualProperty(
                    property_key=prop1,
                    metric_value=transaction_type.MetricValue(value=123.45, unit="USD")
                )
            ],
            calculations=[
                transaction_type.TransactionTypeCalculation(
                    type=transaction_type.CalculationType.TaxAmounts,
                    side=side1,
                    formula="test_formula"
                )
            ],
        )
        # when we dump it
        result = sut.model_dump(
            mode="json",
            by_alias=True,
            round_trip=True,
            exclude_none=True,
            context={"style": "dump"}
        )
        # then property and side references are serialized as $ref
        assert result["scope"] == "sc1"
        assert result["source"] == "default"
        # movement side uses $ref
        assert result["movements"][0]["side"] == {"$ref": "side1"}
        # movement properties use $ref
        assert result["movements"][0]["properties"][0]["propertyKey"] == {"$ref": "prop1"}
        # movement mappings use $ref
        assert result["movements"][0]["mappings"][0]["propertyKey"] == {"$ref": "prop1"}
        # global properties use $ref
        assert result["properties"][0]["propertyKey"] == {"$ref": "prop1"}
        # calculations side uses $ref
        assert result["calculations"][0]["side"] == {"$ref": "side1"}

    def test_undump_with_refs(self):
        # given dump data with $ref values
        prop1 = property.DefinitionRef(
            id="prop1", domain=property.Domain.Transaction, scope="sc1", code="TestProp"
        )
        side1 = side_definition.SideRef(
            id="side1", side="TestSide", scope="sc1"
        )
        data = {
            "scope": "sc1",
            "source": "default",
            "aliases": [
                {
                    "type": "Buy",
                    "description": "Test transaction",
                    "transaction_class": "default",
                    "transaction_roles": "LongLonger",
                    "is_default": False,
                }
            ],
            "movements": [
                {
                    "movement_types": "StockMovement",
                    "side": {"$ref": "side1"},
                    "direction": 1,
                    "properties": [
                        {
                            "property_key": {"$ref": "prop1"},
                            "label_value": "Test Label"
                        }
                    ],
                    "mappings": [
                        {
                            "property_key": {"$ref": "prop1"},
                            "set_to": "Test Value"
                        }
                    ],
                    "name": "Stock Movement",
                    "movement_options": ["DirectAdjustment"],
                    "condition": "",
                }
            ],
            "properties": [
                {
                    "property_key": {"$ref": "prop1"},
                    "metric_value": {"value": 123.45, "unit": "USD"}
                }
            ],
            "calculations": [
                {
                    "type": "TaxAmounts",
                    "side": {"$ref": "side1"},
                    "formula": "test_formula"
                }
            ],
        }
        # when we undump it with $refs context
        result = transaction_type.TransactionTypeResource.model_validate(
            data, context={
                "style": "undump",
                "id": "txn_with_refs",
                "$refs": {
                    "prop1": prop1,
                    "side1": side1,
                }
            }
        )
        # then references are resolved correctly
        assert result.id == "txn_with_refs"
        assert result.scope == "sc1"
        assert result.source == "default"
        # verify movement side is resolved
        movement = result.movements[0]
        assert movement.side == side1
        assert movement.side.side == "TestSide"
        # verify movement properties are resolved
        assert movement.properties
        assert len(movement.properties) == 1
        assert movement.properties[0].property_key == prop1
        assert movement.properties[0].property_key.code == "TestProp"
        assert movement.properties[0].label_value == "Test Label"
        # verify movement mappings are resolved
        assert movement.mappings
        assert len(movement.mappings) == 1
        assert movement.mappings[0].property_key == prop1
        assert movement.mappings[0].set_to == "Test Value"
        # verify global properties are resolved
        assert result.properties
        assert len(result.properties) == 1
        assert result.properties[0].property_key == prop1
        assert result.properties[0].metric_value
        assert result.properties[0].metric_value.value == 123.45
        # verify calculations side is resolved
        assert result.calculations
        assert len(result.calculations) == 1
        assert result.calculations[0].side == side1
        assert result.calculations[0].formula == "test_formula"

    def test_movement_options_accepts_strings(self):
        """Test that movement_options can accept a list of strings like ["Virtual"]"""
        movement = transaction_type.TransactionTypeMovement(
            movement_types=transaction_type.MovementType.StockMovement,
            side=side_definition.SideResource(
                id="side1",
                side="Side1",
                scope="sc1",
                security="Txn:LusidInstrumentId",
                currency="Txn:SettlementCurrency",
                rate="Txn:TradeToPortfolioRate",
                units="Txn:Units",
                amount="Txn:TotalConsideration",
                notional_amount="0",
            ),
            direction=1,
            movement_options=["Virtual"],
        )
        assert movement.movement_options == ["Virtual"]

    def test_movement_options_accepts_enums(self):
        """Test that movement_options can accept a list of MovementOption enums"""
        movement = transaction_type.TransactionTypeMovement(
            movement_types=transaction_type.MovementType.StockMovement,
            side=side_definition.SideResource(
                id="side1",
                side="Side1",
                scope="sc1",
                security="Txn:LusidInstrumentId",
                currency="Txn:SettlementCurrency",
                rate="Txn:TradeToPortfolioRate",
                units="Txn:Units",
                amount="Txn:TotalConsideration",
                notional_amount="0",
            ),
            direction=1,
            movement_options=[
                transaction_type.MovementOption.Virtual,
                transaction_type.MovementOption.Income,
            ],
        )
        assert movement.movement_options == [
            transaction_type.MovementOption.Virtual,
            transaction_type.MovementOption.Income,
        ]




import os
from types import SimpleNamespace

from pytest import fixture

import fbnconfig
from fbnconfig import cleardown_module as cdm
from fbnconfig import fund_accounting as fa
from tests.integration.generate_test_name import gen

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise (
        RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
    )

lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]
host_vars = {}
client = fbnconfig.create_client(lusid_env, token)
base_url = "api/api/chartofaccounts"


@fixture(scope="module")
def setup_deployment():
    scope = gen("cleardown_module")
    yield SimpleNamespace(name=scope)

    items_to_delete = [
        f"{base_url}/{scope}/cdm_int_test_code/cleardownmodules/int_test_cleardown_code",
        f"{base_url}/{scope}/cdm_int_test_code/cleardownmodules/int_test_fbnconfig_cdm_no_rules",
        f"{base_url}/{scope}/cdm_int_test_code/accounts/int_test_account_code",
        f"{base_url}/chartofaccounts/{scope}/cdm_int_test_code"
    ]

    for item in items_to_delete:
        try:
            client.delete(item)
        except Exception as e:
            print(f"Error deleting {item}: {e}")


@fixture
def updated_resources(setup_deployment):
    scope = setup_deployment.name
    chart_of_account_code = "cdm_int_test_code"

    chart_of_account = fa.ChartOfAccountsResource(
        id="int_test_chart_of_account",
        scope=scope,
        code=chart_of_account_code,
        display_name="fbn config COA integration test"
    )

    account = fa.AccountResource(
        id="int_test_account",
        chart_of_accounts=chart_of_account,
        code="int_test_account_code",
        description="example_desc",
        type=fa.AccountType.ASSET,
        status=fa.AccountStatus.ACTIVE,
        control="Manual"
    )
    # Need to create a chart of account first and general ledger profile is the code for the account

    cdm_resource = cdm.CleardownModuleResource(
        id="int_test_clearndown_module",
        chart_of_accounts=chart_of_account,
        cleardown_module_code="int_test_cleardown_code",
        display_name="updated clear down module display name",
        rules=[cdm.CleardownModuleRule(
            rule_id="rule_0002",
            general_ledger_account_code="int_test_account_code",
            rule_filter="Account.Code startswith '300'",
        )]
    )

    cdm_no_rules_resource = cdm.CleardownModuleResource(
        id="int_test_cleardown_module_no_rules",
        chart_of_accounts=chart_of_account,
        cleardown_module_code="int_test_fbnconfig_cdm_no_rules",
        display_name="clear down module with no rules",
        rules=None
    )

    return [chart_of_account, account, cdm_resource, cdm_no_rules_resource]


@fixture
def base_resources(setup_deployment):
    scope = setup_deployment.name
    chart_of_account_code = "cdm_int_test_code"

    chart_of_account = fa.ChartOfAccountsResource(
        id="int_test_chart_of_account",
        scope=scope,
        code=chart_of_account_code,
        display_name="fbn config COA integration test"
    )

    account = fa.AccountResource(
        id="int_test_account",
        chart_of_accounts=chart_of_account,
        code="int_test_account_code",
        description="example_desc",
        type=fa.AccountType.ASSET,
        status=fa.AccountStatus.ACTIVE,
        control="Manual"
    )
    # Need to create a chart of account first and general ledger profile is the code for the account

    cdm_resource = cdm.CleardownModuleResource(
        id="int_test_clearndown_module",
        chart_of_accounts=chart_of_account,
        cleardown_module_code="int_test_cleardown_code",
        display_name="clear down module code",
        rules=[cdm.CleardownModuleRule(
            rule_id="rule_0001",
            general_ledger_account_code="int_test_account_code",
            rule_filter="Account.Code startswith '200'",
        )]
    )

    cdm_no_rules_resource = cdm.CleardownModuleResource(
        id="int_test_cleardown_module_no_rules",
        chart_of_accounts=chart_of_account,
        cleardown_module_code="int_test_fbnconfig_cdm_no_rules",
        display_name="clear down module with no rules",
        rules=None
    )

    return [chart_of_account, account, cdm_resource, cdm_no_rules_resource]


@fixture
def metadata_only_update_resources(setup_deployment):
    """Resources for testing metadata-only updates (display_name, description)"""
    scope = setup_deployment.name
    chart_of_account_code = "cdm_int_test_code"

    chart_of_account = fa.ChartOfAccountsResource(
        id="int_test_chart_of_account",
        scope=scope,
        code=chart_of_account_code,
        display_name="fbn config COA integration test"
    )

    account = fa.AccountResource(
        id="int_test_account",
        chart_of_accounts=chart_of_account,
        code="int_test_account_code",
        description="example_desc",
        type=fa.AccountType.ASSET,
        status=fa.AccountStatus.ACTIVE,
        control="Manual"
    )

    # Same rules, but different display_name and description
    cdm_resource = cdm.CleardownModuleResource(
        id="int_test_clearndown_module",
        chart_of_accounts=chart_of_account,
        cleardown_module_code="int_test_cleardown_code",
        display_name="METADATA ONLY UPDATE - display name",
        description="Updated description for metadata-only test",
        rules=[cdm.CleardownModuleRule(
            rule_id="rule_0001",  # Same rule as base
            general_ledger_account_code="int_test_account_code",
            rule_filter="Account.Code startswith '200'",
        )]
    )

    cdm_no_rules_resource = cdm.CleardownModuleResource(
        id="int_test_cleardown_module_no_rules",
        chart_of_accounts=chart_of_account,
        cleardown_module_code="int_test_fbnconfig_cdm_no_rules",
        display_name="METADATA ONLY UPDATE - no rules display name",
        description="Updated description for no-rules module",
        rules=None
    )

    return [chart_of_account, account, cdm_resource, cdm_no_rules_resource]


@fixture
def rules_only_update_resources(setup_deployment):
    """Resources for testing rules-only updates"""
    scope = setup_deployment.name
    chart_of_account_code = "cdm_int_test_code"

    chart_of_account = fa.ChartOfAccountsResource(
        id="int_test_chart_of_account",
        scope=scope,
        code=chart_of_account_code,
        display_name="fbn config COA integration test"
    )

    account = fa.AccountResource(
        id="int_test_account",
        chart_of_accounts=chart_of_account,
        code="int_test_account_code",
        description="example_desc",
        type=fa.AccountType.ASSET,
        status=fa.AccountStatus.ACTIVE,
        control="Manual"
    )

    # Same display_name, but different rules
    cdm_resource = cdm.CleardownModuleResource(
        id="int_test_clearndown_module",
        chart_of_accounts=chart_of_account,
        cleardown_module_code="int_test_cleardown_code",
        display_name="clear down module code",  # Same as base
        description=None,  # Same as base (None)
        rules=[
            cdm.CleardownModuleRule(
                rule_id="rule_0001",  # Same rule_id but different filter
                general_ledger_account_code="int_test_account_code",
                rule_filter="Account.Code startswith '400'",  # Different filter
            ),
            cdm.CleardownModuleRule(  # Additional rule
                rule_id="rule_0003",
                general_ledger_account_code="int_test_account_code",
                rule_filter="Account.Code startswith '300'",
            )
        ]
    )

    cdm_no_rules_resource = cdm.CleardownModuleResource(
        id="int_test_cleardown_module_no_rules",
        chart_of_accounts=chart_of_account,
        cleardown_module_code="int_test_fbnconfig_cdm_no_rules",
        display_name="clear down module with no rules",  # Same as base
        description=None,  # Same as base
        rules=None
    )

    return [chart_of_account, account, cdm_resource, cdm_no_rules_resource]


def test_cleardown_module(base_resources, setup_deployment):
    scope = setup_deployment.name

    deployment = fbnconfig.Deployment(setup_deployment.name, base_resources)
    fbnconfig.deploy(deployment, lusid_env, token)

    formatted_url = (
        f"api/api/chartofaccounts/{scope}/cdm_int_test_code/"
        f"cleardownmodules/int_test_cleardown_code"
    )

    response = client.get(formatted_url).json()

    assert response is not None

    expected_rules = [
        {
            "ruleId": "rule_0001",
            "generalLedgerAccountCode": "int_test_account_code",
            "ruleFilter": "Account.Code startswith '200'"
        }
    ]
    assert response["cleardownModuleCode"] == "int_test_cleardown_code"
    assert response["displayName"] == "clear down module code"
    assert len(response["rules"]) == 1
    assert response["rules"] == expected_rules
    assert response["chartOfAccountsId"] == {"scope": scope, "code": "cdm_int_test_code"}

    chart_of_account_code = "cdm_int_test_code"
    no_rules_cdm_code = "int_test_fbnconfig_cdm_no_rules"

    no_cdm_rules_url = (
        f"api/api/chartofaccounts/{scope}/{chart_of_account_code}/"
        f"cleardownmodules/{no_rules_cdm_code}"
    )

    no_rules_response = client.get(no_cdm_rules_url).json()
    assert no_rules_response is not None
    assert no_rules_response["cleardownModuleCode"] == "int_test_fbnconfig_cdm_no_rules"
    assert no_rules_response["displayName"] == "clear down module with no rules"
    assert no_rules_response["chartOfAccountsId"] == {"scope": scope, "code": "cdm_int_test_code"}
    assert no_rules_response["rules"] == []


def test_update_cleardown_module(base_resources, setup_deployment, updated_resources):
    scope = setup_deployment.name
    deployment = fbnconfig.Deployment(setup_deployment.name, base_resources)
    fbnconfig.deploy(deployment, lusid_env, token)

    formatted_url = (
        f"api/api/chartofaccounts/{scope}/cdm_int_test_code/"
        f"cleardownmodules/int_test_cleardown_code"
    )

    response = client.get(formatted_url).json()

    assert response is not None

    expected_rules = [
        {
            "ruleId": "rule_0001",
            "generalLedgerAccountCode": "int_test_account_code",
            "ruleFilter": "Account.Code startswith '200'"
        }
    ]
    assert response["cleardownModuleCode"] == "int_test_cleardown_code"
    assert response["displayName"] == "clear down module code"
    assert len(response["rules"]) == 1
    assert response["rules"] == expected_rules
    assert response["chartOfAccountsId"] == {"scope": scope, "code": "cdm_int_test_code"}

    deployment = fbnconfig.Deployment(setup_deployment.name, updated_resources)
    updated_deployment = fbnconfig.deploy(deployment, lusid_env, token)

    chart_of_accounts_changes = [
        a.change for a in updated_deployment if a.type == "ChartOfAccountsResource"
    ]
    account_changes = [
        a.change for a in updated_deployment if a.type == "AccountResource"
    ]
    cleardown_list_changes = [
        a.change for a in updated_deployment if a.type == "CleardownModuleResource"
    ]

    assert chart_of_accounts_changes == ["nochange"]
    assert account_changes == ["nochange"]
    # cdm_no_rules_resource should be nochange as no rules have been updated
    assert cleardown_list_changes == ["update", "nochange"]

    updated_response = client.get(formatted_url).json()
    assert updated_response is not None
    expected_updated_rules = [
        {
            "ruleId": "rule_0002",
            "generalLedgerAccountCode": "int_test_account_code",
            "ruleFilter": "Account.Code startswith '300'"
        }
    ]
    assert updated_response["cleardownModuleCode"] == "int_test_cleardown_code"
    assert updated_response["displayName"] == "updated clear down module display name"
    assert updated_response["chartOfAccountsId"] == response["chartOfAccountsId"]
    assert len(updated_response["rules"]) == 1
    assert updated_response["rules"] == expected_updated_rules
    assert updated_response["chartOfAccountsId"] == {"scope": scope, "code": "cdm_int_test_code"}


def test_nochange(setup_deployment, base_resources):
    deployment = fbnconfig.Deployment(setup_deployment.name, base_resources)
    fbnconfig.deploy(deployment, lusid_env, token)
    updated_deployment = fbnconfig.deploy(deployment, lusid_env, token)

    chart_of_accounts_changes = [
        a.change for a in updated_deployment if a.type == "ChartOfAccountsResource"
    ]
    account_changes = [
        a.change for a in updated_deployment if a.type == "AccountResource"
    ]
    cleardown_list_changes = [
        a.change for a in updated_deployment if a.type == "CleardownModuleResource"
    ]

    assert chart_of_accounts_changes == ["nochange"]
    assert account_changes == ["nochange"]
    assert cleardown_list_changes == ["nochange", "nochange"]


def test_update_metadata_only_integration(base_resources, setup_deployment,
                                          metadata_only_update_resources):
    """Integration test for metadata-only updates (display_name, description)"""
    scope = setup_deployment.name

    # Deploy base resources first
    deployment = fbnconfig.Deployment(setup_deployment.name, base_resources)
    fbnconfig.deploy(deployment, lusid_env, token)

    # Deploy metadata-only update
    metadata_deployment = fbnconfig.Deployment(setup_deployment.name, metadata_only_update_resources)
    updated_deployment = fbnconfig.deploy(metadata_deployment, lusid_env, token)

    cleardown_list_changes = [
        a.change for a in updated_deployment if a.type == "CleardownModuleResource"
    ]

    # Should be update for the main resource (metadata changed) and nochange for no-rules resource
    assert cleardown_list_changes == ["update", "update"]

    # Verify the changes were applied
    formatted_url = (
        f"api/api/chartofaccounts/{scope}/cdm_int_test_code/"
        f"cleardownmodules/int_test_cleardown_code"
    )

    response = client.get(formatted_url).json()
    assert response["displayName"] == "METADATA ONLY UPDATE - display name"
    assert response["description"] == "Updated description for metadata-only test"

    # Rules should be unchanged
    expected_rules = [
        {
            "ruleId": "rule_0001",
            "generalLedgerAccountCode": "int_test_account_code",
            "ruleFilter": "Account.Code startswith '200'"
        }
    ]
    assert response["rules"] == expected_rules


def test_update_rules_only_integration(base_resources, setup_deployment, rules_only_update_resources):
    """Integration test for rules-only updates"""
    scope = setup_deployment.name

    # Deploy base resources first
    deployment = fbnconfig.Deployment(setup_deployment.name, base_resources)
    fbnconfig.deploy(deployment, lusid_env, token)

    # Deploy rules-only update
    rules_deployment = fbnconfig.Deployment(setup_deployment.name, rules_only_update_resources)
    updated_deployment = fbnconfig.deploy(rules_deployment, lusid_env, token)

    cleardown_list_changes = [
        a.change for a in updated_deployment if a.type == "CleardownModuleResource"
    ]

    # Should be update for the main resource (rules changed) and nochange for no-rules resource
    assert cleardown_list_changes == ["update", "nochange"]

    # Verify the changes were applied
    formatted_url = (
        f"api/api/chartofaccounts/{scope}/cdm_int_test_code/"
        f"cleardownmodules/int_test_cleardown_code"
    )

    response = client.get(formatted_url).json()
    assert response["displayName"] == "clear down module code"  # Unchanged
    assert response.get("description") is None  # Unchanged

    # Rules should be updated
    expected_rules = [
        {
            "ruleId": "rule_0001",
            "generalLedgerAccountCode": "int_test_account_code",
            "ruleFilter": "Account.Code startswith '400'"  # Updated filter
        },
        {
            "ruleId": "rule_0003",
            "generalLedgerAccountCode": "int_test_account_code",
            "ruleFilter": "Account.Code startswith '300'"  # New rule
        }
    ]
    assert len(response["rules"]) == 2
    assert response["rules"] == expected_rules


def test_delete_cleardown_module(base_resources, setup_deployment):
    scope = setup_deployment.name
    deployment = fbnconfig.Deployment(setup_deployment.name, base_resources)
    fbnconfig.deploy(deployment, lusid_env, token)

    resources_without_cleardown_modules = base_resources[:2]
    new_deployment = fbnconfig.Deployment(scope, resources_without_cleardown_modules)
    updated_deployment = fbnconfig.deploy(new_deployment, lusid_env, token)

    cleardown_list_changes = [
        a.change for a in updated_deployment if a.type == "CleardownModuleResource"
    ]

    assert cleardown_list_changes == ["remove", "remove"]



import pathlib

from fbnconfig import Deployment, recipe

"""
An example configuration for defining recipe entities.
The script configures the following entities:
- Configuration Recipe

This example will load recipes through two different
methods, via hardcoding or a JSON file.

More information can be found here:
https://support.lusid.com/knowledgebase/article/KA-01855/
"""


def configure(env):
    configuration_recipe = recipe.RecipeResource(
        id="recipe1",
        scope="sc1",
        code="cd1",
        recipe={
            "market": {
                "marketRules": [
                    {
                        "key": "Fx.CurrencyPair.*",
                        "supplier": "DataScope",
                        "dataScope": "SomeScopeToLookAt",
                        "quoteType": "Rate",
                        "field": "Mid",
                        "priceSource": "",
                        "sourceSystem": "Lusid",
                    }
                ],
                "suppliers": {},
                "options": {
                    "defaultSupplier": "Lusid",
                    "defaultInstrumentCodeType": "LusidInstrumentId",
                    "defaultScope": "default",
                    "attemptToInferMissingFx": False,
                    "calendarScope": "CoppClarkHolidayCalendars",
                    "conventionScope": "Conventions",
                },
                "specificRules": [],
                "groupedMarketRules": [],
            },
            "pricing": {
                "modelRules": [],
                "modelChoice": {},
                "options": {
                    "modelSelection": {"library": "Lusid", "model": "SimpleStatic"},
                    "useInstrumentTypeToDeterminePricer": False,
                    "allowAnyInstrumentsWithSecUidToPriceOffLookup": False,
                    "allowPartiallySuccessfulEvaluation": False,
                    "produceSeparateResultForLinearOtcLegs": False,
                    "enableUseOfCachedUnitResults": False,
                    "windowValuationOnInstrumentStartEnd": False,
                    "removeContingentCashflowsInPaymentDiary": False,
                    "useChildSubHoldingKeysForPortfolioExpansion": False,
                    "validateDomesticAndQuoteCurrenciesAreConsistent": False,
                    "conservedQuantityForLookthroughExpansion": "PV",
                },
                "resultDataRules": [],
            },
            "aggregation": {
                "options": {
                    "useAnsiLikeSyntax": False,
                    "allowPartialEntitlementSuccess": False,
                    "applyIso4217Rounding": False,
                }
            },
            "description": "",
            "holding": {"taxLotLevelHoldings": True},
        },
    )
    configuration_recipe_from_drive = recipe.RecipeResource(
        id="recipe2",
        scope="sc1",
        code="cd2",
        recipe=pathlib.Path(__file__).parent.parent.resolve() / pathlib.Path("./data/recipe.json"),
    )
    return Deployment(
        "recipe_example",
        [configuration_recipe, configuration_recipe_from_drive],
    )



import json
from typing import Dict

import httpx
import pytest

from fbnconfig import http_client, log

log_entity = "deployment"


# region test helpers
def build_field(
    name: str = "aname",
    field_description: str = "somed escription",
    is_collection: bool = False,
    lifetime: str = "Perpetual",
    field_type: str = "String",
    required: bool = True,
) -> Dict:
    """
    Helper for testing different values for a FieldSchema object

    :return: a FieldSchema object for a custom entity

    """

    if is_collection:
        return {
            "name": name,
            "lifetime": lifetime,
            "type": field_type,
            "required": required,
            "description": field_description,
            "collectionType": "Array",
        }

    return {
        "name": name,
        "lifetime": lifetime,
        "type": field_type,
        "required": required,
        "description": field_description,
    }


def build_entity_definition(
    name: str = "Deploy Resource", description: str = "Deploy Resource v2.1", fields=None
) -> Dict:
    """
    Returns a CustomEntityDefinition; Defaults to look like the definition set in `log.py`
    """

    return {
        "displayName": name,
        "description": description,
        "fieldSchema": fields
        if fields is not None
        else [
            build_field("dependencies", "IDs of resource this one depends on", True),
            build_field("deployment", "the deployment this resource is in"),
            build_field("resource", "ID of the resource within the deployment"),
            build_field("resourceType", "Class that manages this resource"),
            build_field("state", "Current state resource within the deployment"),
        ],
    }


# endregion


@pytest.mark.respx(base_url="https://foo.lusid.com")
class DescribeSetup:
    @staticmethod
    def response_hook(response: httpx.Response) -> None:
        if response.is_error:
            response.read()
            response.raise_for_status()

    base_url = "https://foo.lusid.com"

    client = http_client.create_client(lusid_url=base_url, token="token")

    def test_setup_from_scratch(self, respx_mock, capsys):
        respx_mock.post("/api/api/propertydefinitions").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json={
                        "href": "https://example.lusid.com/api/properties/demo-scope/demo-code",
                        "key": "identifier/key/something",
                    },
                )
            ]
        )

        respx_mock.post("/api/api/customentitytypes").mock(
            side_effect=[httpx.Response(200, json={"displayName": "something", "anotherKey": "else"})]
        )
        respx_mock.get(f"/api/api/customentitytypes/~{log_entity}").mock(
            side_effect=[httpx.Response(404, json={"name": "CustomEntityDefinitionNotFound"})]
        )
        client = self.client
        # given nothing exists yet
        # when we call setup
        log.setup(client)
        # then it creates a property definition for the identifier
        identifier = respx_mock.calls[0].request
        assert identifier.method == "POST"
        assert identifier.url == f"{self.base_url}/api/api/propertydefinitions"
        # and a custom entity type with entity typename
        entity = respx_mock.calls.last.request
        assert entity.method == "POST"
        assert entity.url == f"{self.base_url}/api/api/customentitytypes"
        assert json.loads(entity.content)["entityTypeName"] == log_entity

        captured = capsys.readouterr()
        assert (
            captured.out
            == "Created identifier identifier/key/something\nCreated entity definition: something\n"
        )

    def test_setup_when_identifier_exists(self, respx_mock):
        respx_mock.post("/api/api/propertydefinitions").mock(
            side_effect=[httpx.Response(400, json={"name": "PropertyAlreadyExists"})]
        )

        respx_mock.post("/api/api/customentitytypes").mock(
            side_effect=[httpx.Response(200, json={"displayName": "something", "anotherKey": "else"})]
        )
        respx_mock.get(f"/api/api/customentitytypes/~{log_entity}").mock(
            side_effect=[httpx.Response(404, json={"name": "CustomEntityDefinitionNotFound"})]
        )
        client = self.client
        # given the identifier exists but the entity does not
        # when we call setup
        log.setup(client)
        # then it creates a property definition for the identifier and handles the error
        identifier = respx_mock.calls[0].request
        assert identifier.method == "POST"
        assert identifier.url == f"{self.base_url}/api/api/propertydefinitions"
        # and a custom entity type
        entity = respx_mock.calls.last.request
        assert entity.method == "POST"
        assert entity.url == f"{self.base_url}/api/api/customentitytypes"

    def test_setup_when_identifier_call_fails(self, respx_mock):
        respx_mock.post("/api/api/propertydefinitions").mock(
            side_effect=[httpx.Response(400, json={"name": "SomeOtherError"})]
        )
        client = self.client
        # given the identifier exists but the entity does not
        # when we call setup it throws
        with pytest.raises(httpx.HTTPStatusError) as error:
            log.setup(client)
        assert error.value.response.status_code == 400
        assert error.value.response.json()["name"] == "SomeOtherError"

    def test_setup_when_entity_exists_with_same_fields(self, respx_mock):
        respx_mock.post("/api/api/propertydefinitions").mock(
            side_effect=[httpx.Response(400, json={"name": "PropertyAlreadyExists"})]
        )

        respx_mock.get(f"/api/api/customentitytypes/~{log_entity}").mock(
            side_effect=[httpx.Response(200, json=build_entity_definition())]
        )
        client = self.client

        # given the identifier and the entity exist and the entity is upto date
        # when we call setup
        log.setup(client)
        # then it creates a property definition for the identifier (and handles the error)
        identifier = respx_mock.calls[0].request
        assert identifier.method == "POST"
        assert identifier.url == f"{self.base_url}/api/api/propertydefinitions"
        # and it does not update the entity

    @pytest.mark.parametrize(
        "test_input",
        [
            build_entity_definition(description="something else"),
            build_entity_definition(name="something else"),
            build_entity_definition(fields=[]),
            build_entity_definition(fields=[build_field(name="different")]),
            build_entity_definition(fields=[build_field(field_description="different")]),
            build_entity_definition(fields=[build_field(is_collection=False)]),
            build_entity_definition(fields=[build_field(lifetime="Something else")]),
            build_entity_definition(fields=[build_field(field_type="Something else")]),
            build_entity_definition(fields=[build_field(required=False)]),
        ],
    )
    def test_setup_when_entity_exists_with_different_fields(self, respx_mock, test_input):
        respx_mock.post("/api/api/propertydefinitions").mock(
            side_effect=[httpx.Response(400, json={"name": "PropertyAlreadyExists"})]
        )

        respx_mock.put(f"/api/api/customentitytypes/~{log_entity}").mock(
            side_effect=[httpx.Response(200, json={})]
        )
        respx_mock.get(f"/api/api/customentitytypes/~{log_entity}").mock(
            side_effect=[httpx.Response(200, json=test_input)]
        )
        client = self.client
        # given the identifier and the entity exist, but the entity is an old version
        # when we call setup
        log.setup(client)
        # then it creates a property definition for the identifier (and handles the error)
        identifier = respx_mock.calls[0].request
        assert identifier.method == "POST"
        assert identifier.url == f"{self.base_url}/api/api/propertydefinitions"
        # and it updates the entity
        # and a custom entity type
        entity = respx_mock.calls.last.request
        assert entity.method == "PUT"
        assert entity.url == f"{self.base_url}/api/api/customentitytypes/~{log_entity}"

    @pytest.mark.parametrize("status_code", [404, 401, 403])
    def test_setup_when_get_entity_fails(self, respx_mock, status_code):
        respx_mock.post("/api/api/propertydefinitions").mock(
            side_effect=[httpx.Response(400, json={"name": "PropertyAlreadyExists"})]
        )

        respx_mock.get(f"/api/api/customentitytypes/~{log_entity}").mock(
            side_effect=[httpx.Response(status_code, json={"Exception": "somexception"})]
        )

        client = self.client
        with pytest.raises(httpx.HTTPStatusError) as ex:
            log.setup(client)
        assert ex.value.response.status_code == status_code


@pytest.mark.respx(base_url="https://foo.lusid.com")
class DescribeList:
    @staticmethod
    def response_hook(response: httpx.Response) -> None:
        if response.is_error:
            response.read()
            response.raise_for_status()

    base_url = "https://foo.lusid.com"
    client = httpx.Client(base_url=base_url, event_hooks={"response": [response_hook]})

    def test_list_deployments_pages_results(self, respx_mock):
        def get_entry(date, id):
            return {
                "description": "...",
                "displayName": "...",
                "fields": [
                    {"name": "deployment", "value": id},
                    {"name": "resourceType", "value": "FileResource"},
                    {"name": "dependencies", "value": []},
                    {"name": "resource", "value": "something"},
                    {"name": "state", "value": '{"some_variable": "some"}'},
                ],
                "identifiers": [
                    {
                        "identifierScope": log_entity,
                        "identifierType": "resource",
                        "identifierValue": "whatever_one",
                    }
                ],
                "version": {"asAtModified": date},
            }

        # given four entries in the log
        existing_values = [
            get_entry("2021-02-20T15:39:58.1008350+00:00", "4"),
            get_entry("2023-02-20T15:39:58.1008350+00:00", "2"),
            get_entry("2024-02-20T15:39:58.1008350+00:00", "1"),
            get_entry("2022-02-20T15:39:58.1008350+00:00", "3"),
        ]
        # and the api returns them in pages of two
        respx_mock.get(f"/api/api/customentities/~{log_entity}").mock(
            side_effect=[
                httpx.Response(200, json={"values": existing_values[0:2], "nextPage": "2"}),
                httpx.Response(200, json={"values": existing_values[2:]}),
            ]
        )
        # when we list them
        logs = log.list_deployments(self.client)
        # Then all four are returned in order
        assert ["1", "2", "3", "4"] == logs
        # and the second call requests page 2
        request = respx_mock.calls.last.request
        assert request.method == "GET"
        assert request.url.path == f"/api/api/customentities/~{log_entity}"
        assert dict(request.url.params) == {"page": "2"}

    def test_list_when_no_entries(self, respx_mock):
        # given there are not entries
        respx_mock.get(f"/api/api/customentities/~{log_entity}").mock(
            side_effect=[httpx.Response(200, json={"values": []})]
        )
        # when we list them
        logs = log.list_deployments(self.client)
        # we get an empty list
        assert logs == []

    def test_list_deployments_returns_in_order(self, respx_mock):
        def get_entry(date, id):
            return {
                "description": "...",
                "displayName": "...",
                "fields": [
                    {"name": "deployment", "value": id},
                    {"name": "resourceType", "value": "FileResource"},
                    {"name": "dependencies", "value": []},
                    {"name": "resource", "value": "something"},
                    {"name": "state", "value": '{"some_variable": "some"}'},
                ],
                "identifiers": [
                    {
                        "identifierScope": log_entity,
                        "identifierType": "resource",
                        "identifierValue": "whatever_one",
                    }
                ],
                "version": {"asAtModified": date},
            }

        existing_values = [
            get_entry("2021-02-20T15:39:58.1008350+00:00", "4"),
            get_entry("2023-02-20T15:39:58.1008350+00:00", "2"),
            get_entry("2024-02-20T15:39:58.1008350+00:00", "1"),
            get_entry("2022-02-20T15:39:58.1008350+00:00", "3"),
        ]

        respx_mock.get(f"/api/api/customentities/~{log_entity}").mock(
            return_value=httpx.Response(200, json={"values": existing_values})
        )

        # Returned in asAt order
        assert ["1", "2", "3", "4"] == log.list_deployments(self.client)

    def test_list_returns_resources_in_order(self, respx_mock):
        def get_entry(date, id):
            return {
                "description": "...",
                "displayName": "...",
                "fields": [
                    {"name": "deployment", "value": "mydeployment"},
                    {"name": "resourceType", "value": "FileResource"},
                    {"name": "dependencies", "value": []},
                    {"name": "resource", "value": id},
                    {"name": "state", "value": '{"some_variable": "some"}'},
                ],
                "identifiers": [
                    {
                        "identifierScope": log_entity,
                        "identifierType": "resource",
                        "identifierValue": "whatever_one",
                    }
                ],
                "version": {"asAtModified": date},
            }

        # given four log entrues for mydeployment
        existing_values = [
            get_entry("2021-02-20T15:39:58.1008350+00:00", "4"),
            get_entry("2023-02-20T15:39:58.1008350+00:00", "2"),
            get_entry("2024-02-20T15:39:58.1008350+00:00", "1"),
            get_entry("2022-02-20T15:39:58.1008350+00:00", "3"),
        ]
        # and the api returns them in pages of two
        respx_mock.get(f"/api/api/customentities/~{log_entity}").mock(
            side_effect=[
                httpx.Response(200, json={"values": existing_values[0:2], "nextPage": "page2"}),
                httpx.Response(200, json={"values": existing_values[2:]}),
            ]
        )
        # The entries are returned in asAt order
        assert ["1", "2", "3", "4"] == [
            line.resource_id for line in log.list_resources_for_deployment(self.client, "mydeployment")
        ]
        # and the filter was sent on both page requests
        request1 = respx_mock.calls[0].request
        assert dict(request1.url.params) == {"filter": "fields[deployment] eq 'mydeployment'"}
        request2 = respx_mock.calls[1].request
        assert dict(request2.url.params) == {
            "filter": "fields[deployment] eq 'mydeployment'",
            "page": "page2",
        }

    def test_get_returns_resource(self, respx_mock):
        response = {
            "description": "...",
            "displayName": "...",
            "fields": [
                {"name": "deployment", "value": "deployment1"},
                {"name": "resourceType", "value": "FileResource"},
                {"name": "dependencies", "value": []},
                {"name": "resource", "value": "something"},
                {"name": "state", "value": '{"some_variable": "some"}'},
            ],
            "identifiers": [
                {
                    "identifierScope": log_entity,
                    "identifierType": "resource",
                    "identifierValue": "whatever_3",
                }
            ],
            "version": {"asAtModified": "2021-02-20T15:39:58.1008350+00:00"},
        }

        filter = "fields[deployment] eq 'deployment1' and fields[resource] eq 'whatever_3'"
        respx_mock.get(f"/api/api/customentities/~{log_entity}", params={"filter": filter}).mock(
            return_value=httpx.Response(200, json={"values": [response]})
        )

        resource = log.get_resource(self.client, "deployment1", "whatever_3")
        assert len(resource) == 1

    def test_get_when_no_resources(self, respx_mock):
        filter = "fields[deployment] eq 'deployment1' and fields[resource] eq 'whatever_3'"

        respx_mock.get(f"/api/api/customentities/~{log_entity}", params={"filter": filter}).mock(
            return_value=httpx.Response(200, json={"values": []})
        )

        resource = log.get_resource(self.client, "deployment1", "whatever_3")
        assert len(resource) == 0

    def test_get_dependencies_returns_correctly(self, respx_mock):
        response = {
            "description": "...",
            "displayName": "...",
            "fields": [
                {"name": "deployment", "value": "deployment1"},
                {"name": "resourceType", "value": "FileResource"},
                {"name": "dependencies", "value": ["a", "b", "c"]},
                {"name": "resource", "value": "whatever_1"},
                {"name": "state", "value": '{"some_variable": "some"}'},
            ],
            "identifiers": [
                {
                    "identifierScope": log_entity,
                    "identifierType": "resource",
                    "identifierValue": "deployment1_whatever_1",
                }
            ],
            "version": {"asAtModified": "2021-02-20T15:39:58.1008350+00:00"},
        }
        respx_mock.get(
            f"/api/api/customentities/~{log_entity}",
            params={"filter": "fields[deployment] eq 'deployment1'"},
        ).mock(return_value=httpx.Response(200, json={"values": [response]}))

        _, resource = log.get_dependencies_map(self.client, "deployment1")
        assert resource["whatever_1"] == ["a", "b", "c"]

    def test_get_dependent_returns_correctly(self, respx_mock):
        def get_entry(deployment, id, deps):
            return {
                "description": "...",
                "displayName": "...",
                "fields": [
                    {"name": "deployment", "value": deployment},
                    {"name": "resourceType", "value": "FileResource"},
                    {"name": "dependencies", "value": deps},
                    {"name": "resource", "value": id},
                    {"name": "state", "value": '{"some_variable": "some"}'},
                ],
                "identifiers": [
                    {
                        "identifierScope": log_entity,
                        "identifierType": "resource",
                        "identifierValue": id,
                    }
                ],
                "version": {"asAtModified": "2021-02-20T15:39:58.1008350+00:00"},
            }

        # A -> B,C
        # B -> E
        # D -> F
        # E <- B
        # F <- D
        # Z no deps

        respx_mock.get(
            f"/api/api/customentities/~{log_entity}",
            params={"filter": "fields[deployment] eq 'deployment1'"},
        ).mock(
            return_value=httpx.Response(
                200,
                json={
                    "values": [
                        get_entry("deployment1", "a", ["b", "c"]),
                        get_entry("deployment1", "b", ["e"]),
                        get_entry("deployment1", "d", ["f"]),
                        get_entry("deployment1", "f", []),
                        get_entry("deployment1", "e", []),
                        get_entry("deployment1", "z", []),
                    ]
                },
            )
        )

        _, uses_map = log.get_dependents_map(self.client, "deployment1")
        assert uses_map["e"] == ["b"]
        assert uses_map["f"] == ["d"]
        assert uses_map["z"] == []

    @pytest.mark.parametrize(
        "deployment_id,resource_id",
        [("clean", "clean"), ("dir/ty", "clean"), ("clean", "dir/ty"), ("dir/ty", "dir/ty")],
    )
    def test_remove_replaces_slashes_url(self, respx_mock, deployment_id, resource_id):
        client = self.client

        identifier = f"{deployment_id}_{resource_id}".replace("/", "_")
        respx_mock.delete(
            f"/api/api/customentities/~{log_entity}/resource/{identifier}",
            params={"identifierScope": log_entity},
        ).mock(return_value=httpx.Response(200, json={"asAt": "2018-03-05T10:10:10.0000000+00:00"}))

        response = log.remove(client, deployment_id, resource_id)

        assert response == {"asAt": "2018-03-05T10:10:10.0000000+00:00"}

        request = respx_mock.calls.last.request
        assert request.url.path == f"/api/api/customentities/~{log_entity}/resource/{identifier}"



import os
from types import SimpleNamespace

from httpx import HTTPStatusError
from pytest import fixture

import fbnconfig
from fbnconfig import Deployment, fund_accounting, posting_module, property
from tests.integration.generate_test_name import gen

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise (
        RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
    )

lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]
host_vars = {}


def configure(env):
    deployment_name = getattr(env, "name", "chart_of_accounts")

    perp_prop = fund_accounting.PropertyValue(
        property_key=property.DefinitionRef(
            id="one", domain=property.Domain.ChartOfAccounts, scope="sc1", code="cd1"
        ),
        label_value="Hello"
    )

    chart_of_account = fund_accounting.ChartOfAccountsResource(
        id="example_id",
        scope=f"testScope-{deployment_name}",
        code="testCode",
        display_name="example_display_name",
        description="example_description",
        properties=[perp_prop]
    )

    perp_prop_two = fund_accounting.PropertyValue(
        property_key=property.DefinitionRef(
            id="two", domain=property.Domain.Account, scope="sc2", code="cd2"
        ),
        label_value="Goodbye"
    )

    account = fund_accounting.AccountResource(
        id=f"example_id-{deployment_name}",
        chart_of_accounts=chart_of_account,
        code=f"account_code-{deployment_name}",
        description="example_desc",
        type=fund_accounting.AccountType.ASSET,
        status=fund_accounting.AccountStatus.ACTIVE,
        control="Manual",
        properties=[perp_prop_two]
    )

    rule = posting_module.PostingModuleRule(
        rule_id="rule_id",
        general_ledger_account_code=account,
        rule_filter="SourceType eq 'LusidTransaction'"
    )

    post_mod = posting_module.PostingModuleResource(
        id="posting_module_id",
        chart_of_accounts=chart_of_account,
        code=f"module_code-{deployment_name}",
        display_name="example_display_name",
        description="example_description",
        rules=[rule]
    )

    return Deployment(deployment_name, [account, chart_of_account, post_mod])


@fixture(scope="module")
def setup_deployment():
    print("Creating new property")
    client = fbnconfig.create_client(lusid_env, token)

    # Deletes test property for chart of account if it exsits
    try:
        client.delete("/api/api/propertydefinitions/ChartOfAccounts/sc1/cd1")
    except Exception:
        pass
    # Deletes test property for account if it exsits
    try:
        client.delete("/api/api/propertydefinitions/Account/sc2/cd2")
    except Exception:
        pass

    # Creates new property to add to the chart of accounts
    chart_property = {
        "domain": "ChartOfAccounts",
        "scope": "sc1",
        "code": "cd1",
        "valueRequired": False,
        "displayName": "My Property Display Name",
        "dataTypeId": {
            "scope": "system",
            "code": "string"
        },
        "lifeTime": "Perpetual",
    }
    client.post("/api/api/propertydefinitions", json=chart_property)

    # Creates new property add to the chart of accounts
    account_property = {
        "domain": "Account",
        "scope": "sc2",
        "code": "cd2",
        "valueRequired": False,
        "displayName": "My Property Display Name",
        "dataTypeId": {
            "scope": "system",
            "code": "string"
        },
        "lifeTime": "Perpetual",
    }
    client.post("/api/api/propertydefinitions", json=account_property)

    deployment_name = gen("chart_of_accounts")
    print(f"\nRunning for deployment {deployment_name}...")

    yield SimpleNamespace(name=deployment_name)
    # Teardown: Clean up resources (if any) after the test
    print("\nTearing down resources...")
    fbnconfig.deployex(fbnconfig.Deployment(deployment_name, []), lusid_env, token)
    client.delete("/api/api/propertydefinitions/ChartOfAccounts/sc1/cd1")
    client.delete("/api/api/propertydefinitions/Account/sc2/cd2")


def test_teardown(setup_deployment):
    deployment_name = setup_deployment.name
    client = fbnconfig.create_client(lusid_env, token)
    # setup deployment
    fbnconfig.deployex(configure(setup_deployment), lusid_env, token)

    chart_url = f"/api/api/chartofaccounts/testScope-{deployment_name}/testCode"
    account_url = chart_url + f"/accounts/account_code-{deployment_name}"
    posting_mod_url = chart_url + f"/postingmodules/module_code-{deployment_name}"

    # check it exists
    client.request("get", chart_url)
    client.request("get", account_url)
    client.request("get", posting_mod_url)
    # Tear it down
    fbnconfig.deployex(fbnconfig.Deployment(deployment_name, []), lusid_env, token)

    # check account was deleted
    try:
        client.request("get", account_url)
    except HTTPStatusError as error:
        assert error.response.status_code == 404
    # check chart was deleted
    try:
        client.request("get", posting_mod_url)
    except HTTPStatusError as error:
        assert error.response.status_code == 404
    # check chart was deleted
    try:
        client.request("get", chart_url)
    except HTTPStatusError as error:
        assert error.response.status_code == 404


def test_create(setup_deployment):
    deployment_name = setup_deployment.name
    fbnconfig.deployex(configure(setup_deployment), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token)
    chart_url = f"/api/api/chartofaccounts/testScope-{deployment_name}/testCode"
    search_chart = client.request("get", chart_url)
    account_url = chart_url + f"/accounts/account_code-{deployment_name}"
    search_account = client.request("get", account_url)
    assert search_chart.status_code == 200
    assert search_account.status_code == 200


def test_update_nochange(setup_deployment):
    fbnconfig.deployex(configure(setup_deployment), lusid_env, token)
    update = fbnconfig.deployex(configure(setup_deployment), lusid_env, token)
    assert [a.change for a in update if a.type == "ChartOfAccountsResource"] == ["nochange"]
    assert [a.change for a in update if a.type == "AccountResource"] == ["nochange"]
    assert [a.change for a in update if a.type == "PostingModuleResource"] == ["nochange"]



import json
import string
from hashlib import sha256
from typing import Any, Dict, List

import httpx
from pydantic import BaseModel, Field, model_validator

from fbnconfig import property as prop

from .resource_abc import CamelAlias, Resource, register_resource
from .urlfmt import Urlfmt

url: string.Formatter = Urlfmt("api/api/instrumenteventtypes")

TRANSACTION_TEMPLATE_URL_PATTERN = (
    "{base}/{instrumentEventType}/transactiontemplates/{instrumentType}/{scope}"
)


class TransactionCurrencyAndAmount(CamelAlias, BaseModel):
    currency: str | None = None
    amount: str | None = None


class TransactionPriceAndType(CamelAlias, BaseModel):
    price: str | None = None
    type: str | None = None


class TransactionFieldMap(CamelAlias, BaseModel):
    instrument: str
    settlement_date: str
    source: str
    transaction_currency: str
    transaction_date: str
    transaction_id: str
    type: str
    units: str
    transaction_price: TransactionPriceAndType
    exchange_rate: str | None
    total_consideration: TransactionCurrencyAndAmount


class TransactionPropertyMap(CamelAlias, BaseModel):
    property_key: prop.PropertyKey | None
    value: str | None = None


class ComponentTransactions(CamelAlias, BaseModel):
    display_name: str
    condition: str | None = None
    transaction_field_map: TransactionFieldMap
    transaction_property_map: List[TransactionPropertyMap]
    preserve_tax_lot_structure: bool | None = None
    market_open_time_adjustments: str | None = None


@register_resource()
class TransactionTemplateResource(BaseModel, Resource):
    id: str = Field(exclude=True)
    scope: str
    instrument_type: str
    instrument_event_type: str
    description: str
    component_transactions: List[ComponentTransactions]

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data, info) -> Dict[str, Any]:
        if not isinstance(data, dict):
            return data
        # Handle id from context (for dump/undump)
        if info.context and info.context.get("id"):
            data = data | {"id": info.context.get("id")}
        return data

    def read(self, client: httpx.Client, old_state):
        scope = old_state.scope
        instrument_event_type = old_state.instrumentEventType
        instrument_type = old_state.instrumentType

        formatted_url = url.format(
            TRANSACTION_TEMPLATE_URL_PATTERN,
            scope=scope,
            instrumentEventType=instrument_event_type,
            instrumentType=instrument_type
        )
        return client.get(formatted_url).json()

    def create(self, client: httpx.Client):
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        scope = desired["scope"]
        instrument_event_type = desired["instrumentEventType"]
        instrument_type = desired["instrumentType"]
        sorted_desired = json.dumps(desired, sort_keys=True)
        content_hash = sha256(sorted_desired.encode()).hexdigest()
        formatted_url = url.format(
            TRANSACTION_TEMPLATE_URL_PATTERN,
            scope=scope,
            instrumentEventType=instrument_event_type,
            instrumentType=instrument_type
        )

        client.post(formatted_url, json=desired)

        return {
            "id": self.id,
            "scope": self.scope,
            "instrumentType": self.instrument_type,
            "instrumentEventType": self.instrument_event_type,
            "content_hash": content_hash
        }

    @staticmethod
    def delete(client: httpx.Client, old_state):
        scope = old_state.scope
        instrument_event_type = old_state.instrumentEventType
        instrument_type = old_state.instrumentType

        formatted_url = url.format(
            TRANSACTION_TEMPLATE_URL_PATTERN,
            scope=scope,
            instrumentEventType=instrument_event_type,
            instrumentType=instrument_type
        )
        client.delete(formatted_url)

    def update(self, client: httpx.Client, old_state):
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        sorted_desired = json.dumps(desired, sort_keys=True)

        content_hash = sha256(sorted_desired.encode()).hexdigest()

        if (self.scope, self.instrument_type, self.instrument_event_type) != (
            old_state.scope,
            old_state.instrumentType,
            old_state.instrumentEventType
        ):
            self.delete(client, old_state)
            return self.create(client)

        if content_hash == old_state.content_hash:
            return None

        self.delete(client, old_state)
        return self.create(client)

    def deps(self):
        property_keys = []

        for tx in self.component_transactions:
            for prop_map in tx.transaction_property_map:
                if prop_map.property_key is not None:
                    property_keys.append(prop_map.property_key)

        return property_keys




import datetime as dt
import json
from types import SimpleNamespace
from zoneinfo import ZoneInfo

import httpx
import pytest

from fbnconfig import identifier_definition, property

TEST_BASE = "https://foo.lusid.com"


@pytest.mark.respx(base_url=TEST_BASE)
class TestIdentifierDefinitionRef:
    """Test IdentifierDefinitionRef functionality."""

    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [lambda x: x.raise_for_status()]})

    @pytest.fixture
    def identifier_definition_ref(self):
        """Create a identifier definition reference for testing."""
        return identifier_definition.IdentifierDefinitionRef(
            id="identDefId",
            domain=identifier_definition.SupportedDomain.Instrument,
            identifier_scope="testScope",
            identifier_type="testType"
        )

    def test_identifier_def_ref_attach_success(self, respx_mock, identifier_definition_ref):
        respx_mock.get("/api/api/identifierdefinitions/Instrument/testScope/testType").mock(
            return_value=httpx.Response(200, json={})
        )
        # Should not raise an error
        identifier_definition_ref.attach(self.client)

    def test_identifier_def_ref_attach_not_found(self, respx_mock, identifier_definition_ref):
        respx_mock.get("/api/api/identifierdefinitions/Instrument/testScope/testType").mock(
            return_value=httpx.Response(404)
        )
        with pytest.raises(RuntimeError) as ex:
            identifier_definition_ref.attach(self.client)
        assert "Identifier Definition Instrument/testScope/testType does not exist" in str(ex.value)

    def test_identifier_def_ref_attach_when_http_error(self, respx_mock, identifier_definition_ref):
        respx_mock.get("/api/api/identifierdefinitions/Instrument/testScope/testType").mock(
            return_value=httpx.Response(500, json={})
        )
        client = self.client
        sut = identifier_definition_ref
        # when we call attach an exception is raised
        with pytest.raises(httpx.HTTPError):
            sut.attach(client)


@pytest.mark.respx(base_url=TEST_BASE)
class TestIdentifierDefinitionResource:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [lambda x: x.raise_for_status()]})

    @pytest.fixture
    def simple_identifier(self):
        property_example = property.DefinitionRef(
            id="property_id", domain=property.Domain.Instrument, scope="sc1", code="cd4"
        )

        perp_prop = identifier_definition.PropertyValue(
            property_key=property_example,
            label_value="Hello"
        )

        return identifier_definition.IdentifierDefinitionResource(
            id="identDefId",
            domain=identifier_definition.SupportedDomain.Instrument,
            identifier_scope="testScope",
            identifier_type="testType",
            life_time=property.LifeTime.Perpetual,
            hierarchy_usage="ParentIdentifier",
            hierarchy_level="hierarchyLevelExample",
            display_name="example_display_name",
            description="example_description",
            properties=[perp_prop]
        )

    def test_read_ident_def(self, respx_mock, simple_identifier):
        response = {
            "href": "http://example.lusid.com/api/identifierdefinitions/Instrument/oldScope/oldType",
            "domain": "Instrument",
            "identifierScope": "oldScope",
            "identifierType": "oldType",
            "lifeTime": "Perpetual",
            "hierarchyUsage": "MasterIdentifer",
            "hierarchyLevel": "Exchange",
            "displayName": "My Identifier Definition",
            "description": "Optional Identifier definition description",
            "properties": {},
            "version": {
                "effectiveFrom": "2024-06-01T10:30:00.0000000+00:00",
                "asAtDate": "2024-06-01T10:30:00.0000000+00:00",
                "asAtCreated": "2024-06-01T10:30:00.0000000+00:00",
                "userIdCreated": "User1",
                "requestIdCreated": "RequestId1",
                "reasonCreated": "",
                "asAtModified": "2024-06-04T10:30:00.0000000+00:00",
                "userIdModified": "User2",
                "requestIdModified": "RequestId2",
                "reasonModified": "",
                "asAtVersionNumber": 2,
                "entityUniqueId": "00000000-0000-0000-0000-000000000000"
            }
        }

        respx_mock.get(
            "/api/api/identifierdefinitions/Instrument/oldScope/oldType"
        ).mock(return_value=httpx.Response(200, json=response))

        client = self.client
        old_state = SimpleNamespace(
            domain="Instrument",
            identifier_scope="oldScope",
            identifier_type="oldType"
        )
        actual_response = simple_identifier.read(client, old_state)

        # Actual response should have popped href
        assert actual_response != response

        # Simulate pop for response and now assert they're the same
        response.pop("href", None)
        assert actual_response == response

    def test_create_ident_def(self, respx_mock, simple_identifier):
        respx_mock.post("/api/api/identifierdefinitions").mock(
            return_value=httpx.Response(
                200,
                json={
                    "version": {
                        "asAtVersionNumber": 2,
                    }
                }
            )
        )
        client = self.client
        state = simple_identifier.create(client)
        req = respx_mock.calls.last.request
        body = json.loads(req.content)

        # Verify request body structure
        assert body == {
            "domain": "Instrument",
            "identifierScope": "testScope",
            "identifierType": "testType",
            "lifeTime": "Perpetual",
            "hierarchyUsage": "ParentIdentifier",
            "hierarchyLevel": "hierarchyLevelExample",
            "displayName": "example_display_name",
            "description": "example_description",
            "properties": {
                "Instrument/sc1/cd4": {
                    "key": "Instrument/sc1/cd4",
                    "value": {
                        "labelValue": "Hello"
                    },
                }
            }
        }

        source_version = simple_identifier.__get_content_hash__()

        # Verify returned state
        assert state == {
            "domain": "Instrument",
            "identifier_scope": "testScope",
            "identifier_type": "testType",
            "source_version": source_version,
            "remote_version": 2
        }

    def test_create_ident_def_prop_with_time_var(self, respx_mock):
        respx_mock.post("/api/api/identifierdefinitions").mock(
            return_value=httpx.Response(
                200,
                json={
                    "version": {
                        "asAtVersionNumber": 2,
                    }
                }
            )
        )
        client = self.client
        property_example = property.DefinitionRef(
            id="property_id", domain=property.Domain.Instrument, scope="sc1", code="cd4"
        )

        effective_from = dt.datetime(2000, 10, 2, 14, 30, 45, tzinfo=ZoneInfo("Europe/London"))
        effective_until = dt.datetime(2030, 10, 2, 14, 30, 45, tzinfo=ZoneInfo("Europe/London"))

        time_prop = identifier_definition.PropertyValue(
            property_key=property_example,
            label_value="Hello",
            effective_from=effective_from,
            effective_until=effective_until
        )

        sut = identifier_definition.IdentifierDefinitionResource(
            id="identDefId",
            domain=identifier_definition.SupportedDomain.Instrument,
            identifier_scope="testScope",
            identifier_type="testType",
            life_time=property.LifeTime.Perpetual,
            hierarchy_usage="ParentIdentifier",
            hierarchy_level="hierarchyLevelExample",
            display_name="example_display_name",
            description="example_description",
            properties=[time_prop]
        )

        state = sut.create(client)
        req = respx_mock.calls.last.request
        body = json.loads(req.content)

        # Verify request body structure
        assert body == {
            "domain": "Instrument",
            "identifierScope": "testScope",
            "identifierType": "testType",
            "lifeTime": "Perpetual",
            "hierarchyUsage": "ParentIdentifier",
            "hierarchyLevel": "hierarchyLevelExample",
            "displayName": "example_display_name",
            "description": "example_description",
            "properties": {
                "Instrument/sc1/cd4": {
                    "key": "Instrument/sc1/cd4",
                    "value": {
                        "labelValue": "Hello"
                    },
                    "effectiveFrom": "2000-10-02T13:30:45Z",
                    "effectiveUntil": "2030-10-02T13:30:45Z"
                }
            }
        }

        source_version = sut.__get_content_hash__()

        # Verify returned state
        assert state == {
            "domain": "Instrument",
            "identifier_scope": "testScope",
            "identifier_type": "testType",
            "source_version": source_version,
            "remote_version": 2
        }

    def test_create_ident_def_defaults_fields(self, respx_mock):
        respx_mock.post("/api/api/identifierdefinitions").mock(
            return_value=httpx.Response(
                200,
                json={
                    "version": {
                        "asAtVersionNumber": 2,
                    }
                }
            )
        )
        property_example = property.DefinitionRef(
            id="property_id", domain=property.Domain.Instrument, scope="sc1", code="cd4"
        )

        perp_prop = identifier_definition.PropertyValue(
            property_key=property_example,
            label_value="Hello"
        )

        missing_hierarchy_usage_field = identifier_definition.IdentifierDefinitionResource(
            id="identDefId",
            domain=identifier_definition.SupportedDomain.Instrument,
            identifier_scope="testScope",
            identifier_type="testType",
            life_time=property.LifeTime.Perpetual,
            hierarchy_level="hierarchyLevelExample",
            display_name="example_display_name",
            description="example_description",
            properties=[perp_prop]
        )

        client = self.client
        state = missing_hierarchy_usage_field.create(client)
        req = respx_mock.calls.last.request
        body = json.loads(req.content)

        # Verify request body structure - hierarchyUsage is defaulted
        assert body == {
            "domain": "Instrument",
            "identifierScope": "testScope",
            "identifierType": "testType",
            "lifeTime": "Perpetual",
            "hierarchyLevel": "hierarchyLevelExample",
            "displayName": "example_display_name",
            "description": "example_description",
            "properties": {
                "Instrument/sc1/cd4": {
                    "key": "Instrument/sc1/cd4",
                    "value": {
                        "labelValue": "Hello"
                    },
                }
            }
        }

        source_version = missing_hierarchy_usage_field.__get_content_hash__()

        # Verify returned state
        assert state == {
            "domain": "Instrument",
            "identifier_scope": "testScope",
            "identifier_type": "testType",
            "source_version": source_version,
            "remote_version": 2
        }

    def test_create_ident_def_missing_null_fields(self, respx_mock):
        respx_mock.post("/api/api/identifierdefinitions").mock(
            return_value=httpx.Response(
                200,
                json={
                    "version": {
                        "asAtVersionNumber": 2,
                    }
                }
            )
        )
        missing_fields_ident_def = identifier_definition.IdentifierDefinitionResource(
            id="identDefId",
            domain=identifier_definition.SupportedDomain.Instrument,
            identifier_scope="testScope",
            identifier_type="testType",
            life_time=property.LifeTime.Perpetual,
        )

        client = self.client
        state = missing_fields_ident_def.create(client)
        req = respx_mock.calls.last.request
        body = json.loads(req.content)

        # Verify request body structure - hierarchyUsage is defaulted
        assert body == {
            "domain": "Instrument",
            "identifierScope": "testScope",
            "identifierType": "testType",
            "lifeTime": "Perpetual",
        }

        source_version = missing_fields_ident_def.__get_content_hash__()

        # Verify returned state
        assert state == {
            "domain": "Instrument",
            "identifier_scope": "testScope",
            "identifier_type": "testType",
            "source_version": source_version,
            "remote_version": 2
        }

    def test_update_ident_def_without_change(self, respx_mock, simple_identifier):
        respx_mock.get(
            "/api/api/identifierdefinitions/Instrument/"
            f"{simple_identifier.identifier_scope}/{simple_identifier.identifier_type}"
        ).mock(
            return_value=httpx.Response(
                200,
                json={
                    "version": {
                        "asAtVersionNumber": 2,
                    }
                }
            )
        )

        # Set source_version hash to the same to test no change
        source_version = simple_identifier.__get_content_hash__()

        old_state = SimpleNamespace(
            domain="Instrument",
            identifier_scope="testScope",
            identifier_type="testType",
            remote_version=2,
            source_version=source_version
        )

        # Same hash so we expect to return None
        result = simple_identifier.update(self.client, old_state)
        assert result is None

    def test_update_ident_def_with_change(self, respx_mock, simple_identifier):
        respx_mock.get(
            "/api/api/identifierdefinitions/Instrument/"
            "testScope/testType"
        ).mock(
            return_value=httpx.Response(
                200,
                json={
                    "domain": "Instrument",
                    "identifierScope": "testScope",
                    "identifierType": "testType",
                    "lifeTime": "Perpetual",
                    "hierarchyUsage": "MasterIdentifier",
                    "hierarchyLevel": "Exchange",
                    "displayName": "My Identifier Definition",
                    "description": "Optional Identifier definition description",
                    "properties": {},
                    "version": {
                        "effectiveFrom": "2024-06-01T10:30:00.0000000+00:00",
                        "asAtDate": "2024-06-01T10:30:00.0000000+00:00",
                        "asAtCreated": "2024-06-01T10:30:00.0000000+00:00",
                        "userIdCreated": "User1",
                        "requestIdCreated": "RequestId1",
                        "reasonCreated": "",
                        "asAtModified": "2024-06-04T10:30:00.0000000+00:00",
                        "userIdModified": "User2",
                        "requestIdModified": "RequestId2",
                        "reasonModified": "",
                        "asAtVersionNumber": 2,
                        "entityUniqueId": "00000000-0000-0000-0000-000000000000"
                    }
                }
            )
        )
        respx_mock.put(
            "/api/api/identifierdefinitions/Instrument/"
            "testScope/testType"
        ).mock(
            return_value=httpx.Response(
                200,
                json={
                    "version": {
                        "asAtVersionNumber": 2,
                    }
                }
            )
        )

        source_version = simple_identifier.__get_content_hash__()

        old_state = SimpleNamespace(
            domain="Instrument",
            identifier_scope="testScope",
            identifier_type="testType",
            remote_version=123456789,
            source_version=source_version
        )

        # Different hashes so we expect change
        state = simple_identifier.update(self.client, old_state)
        req = respx_mock.calls.last.request
        assert req.method == "PUT"
        body = json.loads(req.content)

        # Verify request body structure
        assert body == {
            "lifeTime": "Perpetual",
            "hierarchyUsage": "ParentIdentifier",
            "hierarchyLevel": "hierarchyLevelExample",
            "displayName": "example_display_name",
            "description": "example_description",
            "properties": {
                "Instrument/sc1/cd4": {
                    "key": "Instrument/sc1/cd4",
                    "value": {
                        "labelValue": "Hello"
                    },
                }
            }
        }

        # Verify returned state
        assert state == {
            "domain": "Instrument",
            "identifier_scope": "testScope",
            "identifier_type": "testType",
            "source_version": source_version,
            "remote_version": 2
        }

    def test_update_ident_def_change_domain(self, respx_mock, simple_identifier):
        respx_mock.delete(
            "/api/api/identifierdefinitions/Person/"
            "testScope/testType"
            ).mock(
            side_effect=[httpx.Response(200, json={})]
        )
        respx_mock.post("/api/api/identifierdefinitions").mock(
            return_value=httpx.Response(
                200,
                json={
                    "version": {
                        "asAtVersionNumber": 2,
                    }
                }
            )
        )

        source_version = simple_identifier.__get_content_hash__()

        #  Domain is different - should call delete, create
        old_state = SimpleNamespace(
            domain="Person",
            identifier_scope="testScope",
            identifier_type="testType",
            remote_version=2,
            source_version=source_version
        )

        state = simple_identifier.update(self.client, old_state)

        #  Assert Delete and Post are called
        assert respx_mock.calls[0].request.method == "DELETE"
        assert respx_mock.calls.last.request.method == "POST"
        req = respx_mock.calls.last.request
        body = json.loads(req.content)

        # Verify request body structure
        assert body == {
            "domain": "Instrument",
            "identifierScope": "testScope",
            "identifierType": "testType",
            "lifeTime": "Perpetual",
            "hierarchyUsage": "ParentIdentifier",
            "hierarchyLevel": "hierarchyLevelExample",
            "displayName": "example_display_name",
            "description": "example_description",
            "properties": {
                "Instrument/sc1/cd4": {
                    "key": "Instrument/sc1/cd4",
                    "value": {
                        "labelValue": "Hello"
                    },
                }
            }
        }

        # Verify returned state
        assert state == {
            "domain": "Instrument",
            "identifier_scope": "testScope",
            "identifier_type": "testType",
            "source_version": source_version,
            "remote_version": 2
        }

    def test_update_ident_def_change_type(self, respx_mock, simple_identifier):
        respx_mock.delete(
            "/api/api/identifierdefinitions/Instrument/"
            "testScope/differentType"
            ).mock(
            side_effect=[httpx.Response(200, json={})]
        )
        respx_mock.post("/api/api/identifierdefinitions").mock(
            return_value=httpx.Response(
                200,
                json={
                    "version": {
                        "asAtVersionNumber": 2,
                    }
                }
            )
        )

        source_version = simple_identifier.__get_content_hash__()

        #  Type is different - should call delete, create
        old_state = SimpleNamespace(
            domain="Instrument",
            identifier_scope="testScope",
            identifier_type="differentType",
            remote_version=2,
            source_version=source_version
        )

        state = simple_identifier.update(self.client, old_state)

        #  Assert Delete and Post are called
        assert respx_mock.calls[0].request.method == "DELETE"
        assert respx_mock.calls.last.request.method == "POST"
        req = respx_mock.calls.last.request
        body = json.loads(req.content)

        # Verify request body structure
        assert body == {
            "domain": "Instrument",
            "identifierScope": "testScope",
            "identifierType": "testType",
            "lifeTime": "Perpetual",
            "hierarchyUsage": "ParentIdentifier",
            "hierarchyLevel": "hierarchyLevelExample",
            "displayName": "example_display_name",
            "description": "example_description",
            "properties": {
                "Instrument/sc1/cd4": {
                    "key": "Instrument/sc1/cd4",
                    "value": {
                        "labelValue": "Hello"
                    },
                }
            }
        }

        # Verify returned state
        assert state == {
            "domain": "Instrument",
            "identifier_scope": "testScope",
            "identifier_type": "testType",
            "source_version": source_version,
            "remote_version": 2
        }

    def test_update_ident_def_change_scope(self, respx_mock, simple_identifier):
        respx_mock.delete(
            "/api/api/identifierdefinitions/Instrument/"
            "oldScope/testType"
            ).mock(
            side_effect=[httpx.Response(200, json={})]
        )
        respx_mock.post("/api/api/identifierdefinitions").mock(
            return_value=httpx.Response(
                200,
                json={
                    "version": {
                        "asAtVersionNumber": 2,
                    }
                }
            )
        )

        source_version = simple_identifier.__get_content_hash__()

        #  Scope is different - should call delete, create
        old_state = SimpleNamespace(
            domain="Instrument",
            identifier_scope="oldScope",
            identifier_type="testType",
            remote_version=2,
            source_version=source_version
        )

        state = simple_identifier.update(self.client, old_state)

        #  Assert Delete and Post are called
        assert respx_mock.calls[0].request.method == "DELETE"
        assert respx_mock.calls.last.request.method == "POST"
        req = respx_mock.calls.last.request
        body = json.loads(req.content)

        # Verify request body structure
        assert body == {
            "domain": "Instrument",
            "identifierScope": "testScope",
            "identifierType": "testType",
            "lifeTime": "Perpetual",
            "hierarchyUsage": "ParentIdentifier",
            "hierarchyLevel": "hierarchyLevelExample",
            "displayName": "example_display_name",
            "description": "example_description",
            "properties": {
                "Instrument/sc1/cd4": {
                    "key": "Instrument/sc1/cd4",
                    "value": {
                        "labelValue": "Hello"
                    },
                }
            }
        }

        # Verify returned state
        assert state == {
            "domain": "Instrument",
            "identifier_scope": "testScope",
            "identifier_type": "testType",
            "source_version": source_version,
            "remote_version": 2
        }

    def test_delete_ident_def(self, respx_mock):
        respx_mock.delete(
            "/api/api/identifierdefinitions/Instrument/"
            "testScope/testType"
            ).mock(
            side_effect=[httpx.Response(200, json={})]
        )
        client = self.client
        old_state = SimpleNamespace(
            domain="Instrument",
            identifier_scope="testScope",
            identifier_type="testType"
        )
        identifier_definition.IdentifierDefinitionResource.delete(client, old_state)
        assert respx_mock.calls.last.request.method == "DELETE"

    def test_deps_without(self, simple_identifier):
        sut = identifier_definition.IdentifierDefinitionResource(
            id="identDefId",
            domain=identifier_definition.SupportedDomain.Instrument,
            identifier_scope="testScope",
            identifier_type="testType",
            life_time=property.LifeTime.Perpetual,
        )
        assert sut.deps() == []

    def test_deps_with_properties(self, simple_identifier):
        sut = simple_identifier
        property_example = property.DefinitionRef(
            id="property_id", domain=property.Domain.Instrument, scope="sc1", code="cd4"
        )
        assert sut.deps() == [property_example]

    def test_dump(self):
        property_example = property.DefinitionRef(
            id="property_id", domain=property.Domain.Instrument, scope="sc1", code="cd4"
        )

        effective_from = dt.datetime(2000, 10, 2, 14, 30, 45, tzinfo=ZoneInfo("Europe/London"))
        effective_until = dt.datetime(2030, 10, 2, 14, 30, 45, tzinfo=ZoneInfo("Europe/London"))

        time_prop = identifier_definition.PropertyValue(
            property_key=property_example,
            label_value="Hello",
            effective_from=effective_from,
            effective_until=effective_until
        )

        sut = identifier_definition.IdentifierDefinitionResource(
            id="identDefId",
            domain=identifier_definition.SupportedDomain.Instrument,
            identifier_scope="testScope",
            identifier_type="testType",
            life_time=property.LifeTime.Perpetual,
            hierarchy_usage="ParentIdentifier",
            hierarchy_level="hierarchyLevelExample",
            display_name="example_display_name",
            description="example_description",
            properties=[time_prop]
        )

        # when we dump it
        result = sut.model_dump(
            mode="json", by_alias=True, round_trip=True, exclude_none=True, context={"style": "dump"}
        )

        assert result == {
            "domain": "Instrument",
            "identifierScope": "testScope",
            "identifierType": "testType",
            "lifeTime": "Perpetual",
            "hierarchyUsage": "ParentIdentifier",
            "hierarchyLevel": "hierarchyLevelExample",
            "displayName": "example_display_name",
            "description": "example_description",
            "properties": [
                {
                    "propertyKey": {
                        "$ref": "property_id"
                    },
                    "labelValue": "Hello",
                    "effectiveFrom": "2000-10-02T13:30:45+00:00",
                    "effectiveUntil": "2030-10-02T13:30:45+00:00"
                }
            ]
        }

    def test_undump(self):
        prop1 = property.DefinitionRef(
            id="property_id", domain=property.Domain.Instrument, scope="sc1", code="TestProp"
        )

        effective_from = dt.datetime(2000, 10, 2, 14, 30, 45, tzinfo=ZoneInfo("Europe/London"))
        effective_until = dt.datetime(2030, 10, 2, 14, 30, 45, tzinfo=ZoneInfo("Europe/London"))

        data = {
            "domain": "Instrument",
            "identifierScope": "testScope",
            "identifierType": "testType",
            "lifeTime": "Perpetual",
            "hierarchyUsage": "ParentIdentifier",
            "hierarchyLevel": "hierarchyLevelExample",
            "displayName": "example_display_name",
            "description": "example_description",
            "properties": [
                {
                    "propertyKey": {
                        "$ref": "property_id"
                    },
                    "labelValue": "Test Label",
                    "effectiveFrom": "2000-10-02T13:30:45+00:00",
                    "effectiveUntil": "2030-10-02T13:30:45+00:00"
                }
            ]
        }

        result = identifier_definition.IdentifierDefinitionResource.model_validate(
            data,
            context={
                "style": "undump",
                "$refs": {
                    "property_id": prop1
                },
                "id": "undump_ident_def",
            }
        )

        assert result.id == "undump_ident_def"
        assert result.domain == identifier_definition.SupportedDomain.Instrument
        assert result.identifier_scope == "testScope"
        assert result.identifier_type == "testType"
        assert result.life_time == "Perpetual"
        assert result.hierarchy_usage == "ParentIdentifier"
        assert result.hierarchy_level == "hierarchyLevelExample"
        assert result.display_name == "example_display_name"
        assert result.description == "example_description"
        assert result.properties
        assert len(result.properties) == 1
        assert result.properties[0].property_key == prop1
        assert result.properties[0].property_key.code == "TestProp"
        assert result.properties[0].label_value == "Test Label"
        assert result.properties[0].effective_from == effective_from
        assert result.properties[0].effective_until == effective_until

    def test_parse_api_format(self):
        prop1 = property.DefinitionRef(
            id="property_id", domain=property.Domain.Instrument, scope="sc1", code="TestProp"
        )

        effective_from = dt.datetime(2000, 10, 2, 14, 30, 45, tzinfo=ZoneInfo("Europe/London"))
        effective_until = dt.datetime(2030, 10, 2, 14, 30, 45, tzinfo=ZoneInfo("Europe/London"))

        data = {
            "domain": "Instrument",
            "identifierScope": "testScope",
            "identifierType": "testType",
            "lifeTime": "Perpetual",
            "hierarchyUsage": "ParentIdentifier",
            "hierarchyLevel": "hierarchyLevelExample",
            "displayName": "example_display_name",
            "description": "example_description",
            "properties": {
                "Instrument/sc1/cd4": {
                    "key": {
                        "$ref": "property_id"
                    },
                    "labelValue": "Test Label",
                    "effectiveFrom": "2000-10-02T13:30:45Z",
                    "effectiveUntil": "2030-10-02T13:30:45Z"
                },
            }
        }

        result = identifier_definition.IdentifierDefinitionResource.model_validate(
            data,
            context={
                "style": "api",
                "$refs": {
                    "property_id": prop1
                },
                "id": "undump_ident_def",
            }
        )

        assert result.id == "undump_ident_def"
        assert result.domain == identifier_definition.SupportedDomain.Instrument
        assert result.identifier_scope == "testScope"
        assert result.identifier_type == "testType"
        assert result.life_time == "Perpetual"
        assert result.hierarchy_usage == "ParentIdentifier"
        assert result.hierarchy_level == "hierarchyLevelExample"
        assert result.display_name == "example_display_name"
        assert result.description == "example_description"
        assert result.properties
        assert len(result.properties) == 1
        assert result.properties[0].property_key == prop1
        assert result.properties[0].property_key.code == "TestProp"
        assert result.properties[0].label_value == "Test Label"
        assert result.properties[0].effective_from == effective_from
        assert result.properties[0].effective_until == effective_until



import json
from typing import Any, Dict

import httpx
import pytest
from pydantic import BaseModel, Field, field_serializer

#
# Not really a test, this shows how to handle the case where
# we want to reference another resource. The user passes in the
# actual resource but when we send we only want to refer to it's
# identifiers.


# Mock/fake resource that will be referenced by another resource
# we pretend that unique_id is a server calculated value
class StubResource(BaseModel):
    id: str = Field(exclude=True)
    scope: str
    code: str
    unique_id: int = Field(None, exclude=True, init=False)

    def read(self):
        self.unique_id = 2345


# a resource that depends on another resource
class UsesAnotherResource(BaseModel):
    id: str = Field(exclude=True)
    # object field which is passed to init but only it's id is sent to the server
    referenced: StubResource = Field(serialization_alias="referencedId")
    _remote: None | Dict[str, Any] = None

    # only send the identifier part of the referenced resource
    @field_serializer("referenced", when_used="always")
    def serialize_job(self, ref: StubResource):
        return ref.unique_id

    def read(self, client, old_state):
        self._remote = client.get("/uses/1").json()

    def create(self, client):
        # use by_alias to call output "referencedId"
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        client.post("/uses", json=desired).json()
        return {}


TEST_BASE = "https://foo.lusid.com"


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeUses:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [lambda r: r.raise_for_status()]})

    def test_create(self, respx_mock):
        respx_mock.post("/uses").mock(return_value=httpx.Response(200, json={}))
        # given a stub resource and a uses
        stub = StubResource(id="stub1", scope="default", code="code1")
        sut = UsesAnotherResource(id="uses1", referenced=stub)
        # and the stub has been (fake) created on the remote
        stub.read()
        # when we create the uses
        sut.create(self.client)
        # then the unique_id of the referenced object gets sent
        # and the alias referenceId gets used as the key
        request = respx_mock.calls.last.request
        assert json.loads(request.content) == {"referencedId": 2345}



from __future__ import annotations

import hashlib
import json
from enum import StrEnum
from types import SimpleNamespace
from typing import Annotated, Any, Dict, Sequence

import httpx
from pydantic import BaseModel, BeforeValidator, Field, PlainSerializer, WithJsonSchema, model_validator

from .resource_abc import Ref, Resource, register_resource


class Visibility(StrEnum):
    PERSONAL = "personal"
    SHARED = "shared"


@register_resource()
class WorkspaceRef(BaseModel, Ref):
    id: str = Field(exclude=True, init=True)
    visibility: Visibility
    name: str

    def attach(self, client):
        try:
            client.request(
                "get", f"/api/api/workspaces/{self.visibility.value}/{self.name}"
            )
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(
                    f"Workspace {self.visibility.value}/{self.name} not found"
                )
            else:
                raise ex


@register_resource()
class WorkspaceResource(BaseModel, Resource):
    """Manage a workspace"""

    id: str = Field(exclude=True)
    visibility: Visibility
    name: str
    description: str

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data: Any, info) -> Dict[str, Any]:
        if not isinstance(data, dict):
            return data
        if info.context and info.context.get("id"):
            return data | {"id": info.context.get("id")}
        return data

    def read(
        self, client: httpx.Client, old_state: SimpleNamespace
    ) -> None | Dict[str, Any]:
        workspace_path = f"{old_state.visibility}/{old_state.name}"
        response = client.get(f"/api/api/workspaces/{workspace_path}").json()
        for key in ["links", "version"]:
            response.pop(key, None)
        return response

    def create(self, client: httpx.Client) -> Dict[str, Any]:
        desired = self.model_dump(mode="json", exclude_none=True, exclude={"visibility"}, by_alias=True)
        sorted_json = json.dumps(desired, sort_keys=True)
        content_hash = hashlib.sha256(sorted_json.encode()).hexdigest()
        client.request("POST", f"/api/api/workspaces/{self.visibility.value}", json=desired)
        return {"visibility": self.visibility, "name": self.name, "content_hash": content_hash}

    def update(self, client: httpx.Client, old_state: SimpleNamespace) -> None | Dict[str, Any]:
        if [old_state.visibility, old_state.name] != [self.visibility, self.name]:
            state = self.create(client)
            self.delete(client, old_state)
            return state
        desired = self.model_dump(mode="json", exclude_none=True, exclude={"visibility"}, by_alias=True)
        sorted_json = json.dumps(desired, sort_keys=True)
        content_hash = hashlib.sha256(sorted_json.encode()).hexdigest()
        if content_hash == getattr(old_state, "content_hash", None):
            return None
        client.request("put", f"/api/api/workspaces/{self.visibility.value}/{self.name}", json=desired)
        return {"visibility": self.visibility, "name": self.name, "content_hash": content_hash}

    @staticmethod
    def delete(client, old_state):
        client.request(
            "delete",
            f"/api/api/workspaces/{old_state.visibility}/{old_state.name}",
        )

    def deps(self) -> Sequence[Resource | Ref]:
        return []


def ser_workspace_key(value, info):
    if info.context and info.context.get("style") == "dump":
        return {"$ref": value.id}
    return value


def des_workspace_key(value, info):
    if not isinstance(value, dict):
        return value
    if info.context and info.context.get("$refs"):
        ref = info.context["$refs"][value["$ref"]]
        return ref
    return value


WorkspaceKey = Annotated[
    WorkspaceResource | WorkspaceRef,
    BeforeValidator(des_workspace_key),
    PlainSerializer(ser_workspace_key),
    WithJsonSchema(
        {
            "type": "object",
            "properties": {"$ref": {"type": "string", "format": "Key.Workspace"}},
            "required": ["$ref"],
        }
    ),
]


@register_resource()
class WorkspaceItemRef(BaseModel, Ref):
    id: str = Field(exclude=True)
    workspace: WorkspaceKey
    group: str
    name: str

    def attach(self, client):
        viz = self.workspace.visibility.value
        name = self.workspace.name
        try:
            client.request(
                "get",
                f"/api/api/workspaces/{viz}/{name}/items/{self.group}/{self.name}",
            )
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(
                    f"Workspace item {viz}/{name}/items/{self.group}/{self.name} not found"
                )
            else:
                raise ex


@register_resource()
class WorkspaceItemResource(BaseModel, Resource):
    id: str = Field(exclude=True)
    workspace: WorkspaceKey
    type: str
    group: str
    name: str
    description: str
    content: Dict[Any, Any]
    format: int

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data: Any, info) -> Dict[str, Any]:
        if not isinstance(data, dict):
            return data
        if info.context and info.context.get("id"):
            return data | {"id": info.context.get("id")}
        return data

    def read(self, client, old_state) -> Dict[str, Any]:
        workspace_path = f"{old_state.visibility}/{old_state.workspace_name}"
        remote = client.request(
            "get",
            f"/api/api/workspaces/{workspace_path}/items/{old_state.group}/{old_state.name}",
        ).json()
        remote.pop("version")
        remote.pop("links")
        return remote

    def create(self, client: httpx.Client):
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True, exclude={"workspace"})
        sorted_json = json.dumps(desired, sort_keys=True)
        content_hash = hashlib.sha256(sorted_json.encode()).hexdigest()
        workspace_path = f"{self.workspace.visibility.value}/{self.workspace.name}"
        client.request(
            "post", f"/api/api/workspaces/{workspace_path}/items", json=desired
        )
        return {
            "visibility": self.workspace.visibility.value,
            "group": self.group,
            "workspace_name": self.workspace.name,
            "name": self.name,
            "content_hash": content_hash,
        }

    def update(self, client: httpx.Client, old_state):
        wksp = self.workspace
        # if we have moved it to a different workspace
        if [old_state.visibility, old_state.workspace_name] != [
            wksp.visibility, wksp.name]:
            state = self.create(client)
            self.delete(client, old_state)
            return state
        # rename the item within the workspace
        if [old_state.name, old_state.group] != [self.name, self.group]:
            state = self.create(client)
            self.delete(client, old_state)
            return state
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True, exclude={"workspace"})
        sorted_json = json.dumps(desired, sort_keys=True)
        content_hash = hashlib.sha256(sorted_json.encode()).hexdigest()
        if content_hash == getattr(old_state, "content_hash", None):
            return None
        workspace_path = f"{self.workspace.visibility.value}/{self.workspace.name}"
        client.request(
            "put",
            f"/api/api/workspaces/{workspace_path}/items/{self.group}/{self.name}",
            json=desired,
        )
        return {
            "visibility": self.workspace.visibility,
            "workspace_name": self.workspace.name,
            "group": self.group,
            "name": self.name,
            "content_hash": content_hash,
        }

    @staticmethod
    def delete(client, old_state):
        workspace_path = f"{old_state.visibility}/{old_state.workspace_name}"
        client.request(
            "delete",
            f"/api/api/workspaces/{workspace_path}/items/{old_state.group}/{old_state.name}",
        )

    def deps(self):
        return [self.workspace]
